"""Package for management communication."""

from __future__ import annotations

import asyncio
from collections.abc import AsyncGenerator, AsyncIterator, Callable, Generator
from contextlib import asynccontextmanager
import logging
import time
from typing import TYPE_CHECKING

from xknx.exceptions import (
    CommunicationError,
    ConfirmationError,
    ManagementConnectionError,
    ManagementConnectionRefused,
    ManagementConnectionTimeout,
)
from xknx.telegram import GroupAddress, IndividualAddress, Telegram
from xknx.telegram.apci import APCI
from xknx.telegram.tpci import (
    TAck,
    TConnect,
    TDataBroadcast,
    TDataConnected,
    TDisconnect,
    TNak,
)
from xknx.util import asyncio_timeout

if TYPE_CHECKING:
    from xknx.xknx import XKNX
logger = logging.getLogger("xknx.management")

MANAGAMENT_ACK_TIMEOUT = 3
MANAGAMENT_CONNECTION_TIMEOUT = 6


class Management:
    """Class for management procedures as described in KNX-Standard 3.5.2."""

    __slots__ = ("_broadcast_contexts", "_connections", "xknx")

    def __init__(self, xknx: XKNX) -> None:
        """Initialize Management class."""
        self.xknx = xknx
        self._connections: dict[IndividualAddress, P2PConnection] = {}
        self._broadcast_contexts: set[BroadcastContext] = set()

    def process(self, telegram: Telegram) -> None:
        """Process incoming telegrams."""
        if isinstance(telegram.tpci, TDataConnected):
            ack = Telegram(
                destination_address=telegram.source_address,
                tpci=TAck(sequence_number=telegram.tpci.sequence_number),
            )
            self.xknx.task_registry.background(
                self.xknx.cemi_handler.send_telegram(ack)
            )
        if conn := self._connections.get(telegram.source_address):
            conn.process(telegram)
            return
        if telegram.tpci.numbered:
            logger.warning(
                "No active point-to-point connection for received telegram: %s",
                telegram,
            )
            return
        if isinstance(telegram.tpci, TConnect):
            # refuse incoming connections
            # TODO: handle incoming telegrams for connections
            # not initiated by us, connection-less and broadcast
            disconnect = Telegram(
                destination_address=telegram.source_address, tpci=TDisconnect()
            )
            self.xknx.task_registry.background(
                self.xknx.cemi_handler.send_telegram(disconnect)
            )
            return
        if isinstance(telegram.tpci, TDataBroadcast):
            for context in self._broadcast_contexts:
                context.queue.put_nowait(telegram)
            return
        logger.debug("Unhandled management telegram: %r", telegram)
        return

    async def connect(
        self, address: IndividualAddress, rate_limit: int = 20
    ) -> P2PConnection:
        """Open a point-to-point connection to a KNX device."""
        if address in self._connections:
            raise ManagementConnectionError(f"Connection to {address} already exists.")
        p2p_connection = P2PConnection(self.xknx, address, rate_limit)
        try:
            await p2p_connection.connect()
        except ManagementConnectionError as exc:
            logger.error("Establishing connection to %s failed: %s", address, exc)
            raise
        self._connections[address] = p2p_connection

        def remove_connection_hook() -> None:
            """Remove connection from management."""
            try:
                del self._connections[address]
            except KeyError:
                logger.error("Connection to %s already closed.", address)

        p2p_connection.disconnect_hook = remove_connection_hook
        return p2p_connection

    async def disconnect(self, address: IndividualAddress) -> None:
        """Close a point-to-point connection to a KNX device."""
        connection = self._connections.get(address)
        if connection is None:
            logger.error(
                "Closing connection to %s failed - connection does not exist.",
                address,
            )
            return
        try:
            await connection.disconnect()
        except ManagementConnectionError as exc:
            logger.error("Closing connection to %s failed: %s", connection.address, exc)
            raise

    @asynccontextmanager
    async def connection(
        self, address: IndividualAddress, rate_limit: int = 20
    ) -> AsyncIterator[P2PConnection]:
        """Provide a point-to-point connection to a KNX device."""
        conn = await self.connect(address, rate_limit)
        try:
            yield conn
        finally:
            await self.disconnect(address)

    async def send_broadcast(self, payload: APCI) -> None:
        """Send a broadcast message."""
        await self.xknx.cemi_handler.send_telegram(
            Telegram(
                GroupAddress("0/0/0"),
                tpci=TDataBroadcast(),
                payload=payload,
            )
        )

    @asynccontextmanager
    async def broadcast(self) -> AsyncIterator[BroadcastContext]:
        """Provide a broadcast context."""
        context = BroadcastContext()
        self._broadcast_contexts.add(context)
        try:
            yield context
        finally:
            self._broadcast_contexts.remove(context)


class BroadcastContext:
    """Class providing broadcast contexts."""

    __slots__ = ("queue",)

    def __init__(self) -> None:
        """Initialize BroadcastContext class."""
        self.queue: asyncio.Queue[Telegram] = asyncio.Queue()

    async def receive(
        self,
        timeout: float | None = 3,
    ) -> AsyncGenerator[Telegram, None]:
        """Receive telegrams from the broadcast context."""
        try:
            async with asyncio_timeout(timeout):
                while True:
                    try:
                        yield await self.queue.get()
                    except GeneratorExit:
                        return
        except asyncio.TimeoutError:
            return


class P2PConnection:
    """Class to manage a point-to-point connection with a KNX device."""

    __slots__ = (
        "_ack_waiter",
        "_connected",
        "_expected_sequence_number",
        "_last_response_time",
        "_response_waiter",
        "address",
        "disconnect_hook",
        "rate_limit",
        "sequence_number",
        "xknx",
    )

    def __init__(
        self, xknx: XKNX, address: IndividualAddress, rate_limit: int = 20
    ) -> None:
        """Initialize P2PConnection class."""
        self.xknx = xknx
        self.address = address
        self.disconnect_hook: Callable[[], None]
        self.rate_limit = rate_limit

        self.sequence_number = self._sequence_number_generator()
        self._expected_sequence_number = 0
        self._connected = False

        self._last_response_time: float = 0

        self._ack_waiter: asyncio.Future[TAck | TNak] | None = None
        self._response_waiter: asyncio.Future[Telegram] = (
            asyncio.get_event_loop().create_future()
        )

    @staticmethod
    def _sequence_number_generator() -> Generator[int, None, None]:
        """Generate sequence numbers."""
        seq_num = 0
        while True:
            yield seq_num
            seq_num = seq_num + 1 & 0xF

    async def connect(self) -> None:
        """Connect to the KNX device."""
        connect = Telegram(
            destination_address=self.address,
            source_address=self.xknx.current_address,
            tpci=TConnect(),
        )
        try:
            await self.xknx.cemi_handler.send_telegram(connect)
        except ConfirmationError as exc:
            self._response_waiter.cancel()
            raise ManagementConnectionError(
                f"Connection to {self.address} failed: {exc}"
            ) from exc
        except CommunicationError as exc:
            self._response_waiter.cancel()
            raise ManagementConnectionError("Error while sending Telegram") from exc
        self._connected = True

    async def disconnect(self) -> None:
        """Disconnect from the KNX device."""
        if not self._connected:
            self.disconnect_hook()  # remove connection from management class
            raise ManagementConnectionRefused(
                "Management connection disconnected by the peer."
            )
        self._connected = False

        disconnect = Telegram(
            destination_address=self.address,
            source_address=self.xknx.current_address,
            tpci=TDisconnect(),
        )
        try:
            await self.xknx.cemi_handler.send_telegram(disconnect)
        except ConfirmationError as exc:
            raise ManagementConnectionError(
                f"Disconnect from {self.address} failed: {exc}"
            ) from exc
        except CommunicationError as exc:
            raise ManagementConnectionError("Error while sending Telegram") from exc
        finally:
            if self._ack_waiter:
                self._ack_waiter.cancel()
            self._response_waiter.cancel()
            self.disconnect_hook()  # remove connection from management class

    def process(self, telegram: Telegram) -> None:
        """Process incoming telegrams."""
        if isinstance(telegram.tpci, TDisconnect):
            logger.info("%s disconnected management session.", self.address)
            self._connected = False
            if self._ack_waiter:
                self._response_waiter.set_exception(ManagementConnectionRefused())
            if not self._response_waiter.done():
                self._response_waiter.set_exception(ManagementConnectionRefused())
            return
        if isinstance(telegram.tpci, TAck | TNak):
            if not self._ack_waiter:
                logger.warning("Received unexpected ACK/NAK: %s", telegram)
                return
            self._ack_waiter.set_result(telegram.tpci)
            return
        if self._response_waiter.done():
            logger.warning(
                "Received unexpected point-to-point telegram for %s: %s",
                self.address,
                telegram,
            )
            return
        if telegram.tpci.sequence_number != self._expected_sequence_number:
            logger.warning(
                "Received unexpected sequence number: %s (expected: %s)",
                telegram.tpci.sequence_number,
                self._expected_sequence_number,
            )
            return
        self._response_waiter.set_result(telegram)
        self._expected_sequence_number = self._expected_sequence_number + 1 & 0xF

    async def _send_data(self, payload: APCI) -> None:
        """
        Send a payload to the KNX device.

        A response has to be processed by `_receive` before sending the next telegram.
        """
        if not self._connected:
            raise ManagementConnectionRefused(
                "Management connection disconnected by the peer."
            )
        self._ack_waiter = asyncio.get_event_loop().create_future()
        seq_num = next(self.sequence_number)
        telegram = Telegram(
            destination_address=self.address,
            source_address=self.xknx.current_address,
            payload=payload,
            tpci=TDataConnected(sequence_number=seq_num),
        )
        try:
            await self.xknx.cemi_handler.send_telegram(telegram)
            async with asyncio_timeout(MANAGAMENT_ACK_TIMEOUT):
                ack = await self._ack_waiter
        except asyncio.TimeoutError:
            logger.info(
                "%s: timeout while waiting for ACK. Resending Telegram.", self.address
            )
            # resend once after 3 seconds without ACK
            # on timeout the Future is cancelled so create a new
            self._ack_waiter = asyncio.get_event_loop().create_future()
            await self.xknx.cemi_handler.send_telegram(telegram)
            try:
                async with asyncio_timeout(MANAGAMENT_ACK_TIMEOUT):
                    ack = await self._ack_waiter
            except asyncio.TimeoutError:
                raise ManagementConnectionTimeout(
                    "No ACK received for repeated telegram."
                ) from None
        except ConfirmationError as exc:
            raise ManagementConnectionError(
                f"Error while sending Telegram: {exc}"
            ) from exc
        except CommunicationError as exc:
            raise ManagementConnectionError("Error while sending Telegram") from exc
        finally:
            self._ack_waiter = None

        if isinstance(ack, TNak):
            raise ManagementConnectionError(
                f"Received no_ack from sending Telegram: {telegram}"
            )
        if ack.sequence_number != seq_num:
            raise ManagementConnectionError(
                f"Ack sequence number {ack.sequence_number} does not match request sequence number of {telegram}"
            )

    async def _receive(self, expected_payload: type[APCI] | None) -> Telegram:
        """Wait for a telegram from the KNX device."""
        try:
            async with asyncio_timeout(MANAGAMENT_CONNECTION_TIMEOUT):
                telegram = await self._response_waiter
        except asyncio.TimeoutError:
            raise ManagementConnectionTimeout(
                f"Timeout while waiting for {expected_payload}"
            ) from None
        finally:
            # set up new Future for the next request
            self._response_waiter = asyncio.get_event_loop().create_future()

        if expected_payload and not isinstance(telegram.payload, expected_payload):
            raise ManagementConnectionError(
                f"Received unexpected telegram: {telegram.payload}"
            )
        return telegram

    async def request(self, payload: APCI, expected: type[APCI] | None) -> Telegram:
        """Send a payload to the KNX device and wait for the response."""
        if not self._connected:
            raise ManagementConnectionRefused(
                "Management connection disconnected by the peer."
            )

        if self.rate_limit:
            # time in seconds since the last request operation
            time_diff = time.time() - self._last_response_time
            wait_time = 1 / self.rate_limit
            if time_diff < wait_time:
                await asyncio.sleep(wait_time - time_diff)

        await self._send_data(payload)
        response = await self._receive(expected)
        self._last_response_time = time.time()
        return response



"""
Abstract base for a specific KNX/IP connection (Tunneling or Routing).

* It handles connection and disconnections
* It starts and stops a udp transport
* It packs Telegrams into KNX Frames and passes them to a udp transport
"""

from __future__ import annotations

from abc import ABC, abstractmethod
from collections.abc import Callable

from xknx.cemi import CEMIFrame

from .transport.ip_transport import KNXIPTransport

CEMIBytesCallbackType = Callable[[bytes], None]


class Interface(ABC):
    """Abstract base class for KNX/IP connections."""

    __slots__ = ("transport",)

    transport: KNXIPTransport

    @abstractmethod
    async def connect(self) -> None:
        """Connect to KNX bus. Raise CommunicationError when not successful."""

    @abstractmethod
    async def disconnect(self) -> None:
        """Disconnect from KNX bus."""

    @abstractmethod
    async def send_cemi(self, cemi: CEMIFrame) -> None:
        """Send CEMIFrame to KNX bus."""



"""Example for Sensor device. See docs/sensor.md and docs/binary_sensor.md for a detailed explanation."""

import asyncio

from xknx import XKNX
from xknx.devices import BinarySensor, Sensor


async def main() -> None:
    """Connect to KNX/IP device and read the value of a temperature and a motion sensor."""
    xknx = XKNX()
    await xknx.start()

    sensor1 = BinarySensor(
        xknx,
        "DiningRoom.Motion.Sensor",
        group_address_state="6/0/2",
    )
    xknx.devices.async_add(sensor1)
    await sensor1.sync(wait_for_result=True)
    print(sensor1)

    sensor2 = Sensor(
        xknx,
        "DiningRoom.Temperature.Sensor",
        group_address_state="6/2/1",
        value_type="temperature",
    )
    xknx.devices.async_add(sensor2)

    await sensor2.sync(wait_for_result=True)
    print(sensor2)

    await xknx.stop()


asyncio.run(main())



"""Example for reading the state from the KNX bus."""

import asyncio

from xknx import XKNX
from xknx.devices import Light


async def main() -> None:
    """Connect to KNX/IP bus and read the state of a Light device."""
    async with XKNX() as xknx:
        light = Light(
            xknx,
            name="TestLight2",
            group_address_switch="1/0/12",
            group_address_brightness="1/0/14",
        )
        xknx.devices.async_add(light)

        await light.set_brightness(80)
        # Will do a GroupValueRead for both addresses and block until a result is received
        await light.sync(wait_for_result=True)
        print(light)


asyncio.run(main())



"""Unit test for KNX/IP RountingIndication objects."""

from xknx.cemi import CEMIFrame, CEMILData, CEMIMessageCode
from xknx.dpt import DPTActiveEnergy
from xknx.knxip import KNXIPFrame, RoutingIndication
from xknx.telegram import GroupAddress, IndividualAddress, Telegram
from xknx.telegram.apci import GroupValueWrite


class TestKNXIPRountingIndication:
    """Class for KNX/IP RoutingIndication test."""

    def test_from_knx(self) -> None:
        """Test parsing and streaming CEMIFrame KNX/IP packet (payload=0xf0)."""
        raw = bytes.fromhex("06 10 05 30 00 12 29 00 bc d0 12 02 01 51 02 00 40 f0")
        knxipframe, _ = KNXIPFrame.from_knx(raw)

        assert isinstance(knxipframe.body, RoutingIndication)
        assert isinstance(knxipframe.body.raw_cemi, bytes)
        assert len(knxipframe.body.raw_cemi) == 12

    def test_from_knx_to_knx(self) -> None:
        """Test parsing and streaming CEMIFrame KNX/IP."""
        raw = bytes.fromhex("06 10 05 30 00 12 29 00 bc d0 12 02 01 51 02 00 40 f0")
        knxipframe, _ = KNXIPFrame.from_knx(raw)

        assert knxipframe.header.to_knx() == raw[:6]
        assert knxipframe.body.to_knx() == raw[6:]
        assert knxipframe.to_knx() == raw

    def test_telegram_set(self) -> None:
        """Test parsing and streaming CEMIFrame KNX/IP packet with DPTArray as payload."""
        telegram = Telegram(
            destination_address=GroupAddress(337),
            payload=GroupValueWrite(
                DPTActiveEnergy().to_knx(0x12345678),
            ),
        )
        cemi = CEMIFrame(
            code=CEMIMessageCode.L_DATA_IND,
            data=CEMILData.init_from_telegram(
                telegram,
                src_addr=IndividualAddress("1.2.2"),
            ),
        )
        assert isinstance(cemi.data, CEMILData)
        cemi.data.hops = 5
        routing_indication = RoutingIndication(raw_cemi=cemi.to_knx())
        knxipframe = KNXIPFrame.init_from_body(routing_indication)

        raw = bytes.fromhex(
            "06 10 05 30 00 15 29 00 bc d0 12 02 01 51 05 00 80 12 34 56 78"
        )

        assert knxipframe.header.to_knx() == raw[:6]
        assert knxipframe.body.to_knx() == raw[6:]
        assert knxipframe.to_knx() == raw

    def test_end_to_end_routing_indication(self) -> None:
        """Test parsing and streaming RoutingIndication KNX/IP packet."""
        # Switch off Kitchen-L1
        raw = bytes.fromhex("06 10 05 30 00 11 29 00 bc d0 ff f9 01 49 01 00 80")
        knxipframe, _ = KNXIPFrame.from_knx(raw)
        raw_cemi = knxipframe.body.raw_cemi

        routing_indication = RoutingIndication(raw_cemi=raw_cemi)
        knxipframe2 = KNXIPFrame.init_from_body(routing_indication)

        assert knxipframe2.header.to_knx() == raw[:6]
        assert knxipframe2.body.to_knx() == raw[6:]
        assert knxipframe2.to_knx() == raw

    def test_from_knx_invalid_cemi(self) -> None:
        """Test parsing and streaming CEMIFrame KNX/IP packet with unsupported CEMICode."""
        routing_indication = RoutingIndication()
        raw = bytes([43, 0, 0, 0, 0, 0, 0, 0, 2, 0, 0])
        assert routing_indication.from_knx(raw) == 11



"""XKNX is an Asynchronous Python module for reading and writing KNX/IP packets."""

from __future__ import annotations

import asyncio
import logging
from logging.handlers import TimedRotatingFileHandler
from pathlib import Path
import signal
from sys import platform
from types import TracebackType

from xknx.cemi import CEMIHandler
from xknx.core import (
    ConnectionManager,
    GroupAddressDPT,
    TaskRegistry,
    TelegramQueue,
)
from xknx.core.state_updater import StateUpdater, TrackerOptionType
from xknx.devices import Device, Devices
from xknx.io import (
    DEFAULT_MCAST_GRP,
    DEFAULT_MCAST_PORT,
    ConnectionConfig,
    knx_interface_factory,
)
from xknx.management import Management
from xknx.telegram import GroupAddress, GroupAddressType, IndividualAddress, Telegram
from xknx.typing import (
    ConnectionChangeCallbackType,
    DeviceCallbackType,
    TelegramCallbackType,
)

from .__version__ import __version__ as VERSION

logger = logging.getLogger("xknx.log")


class XKNX:
    """Class for reading and writing KNX/IP packets."""

    __slots__ = (
        "cemi_handler",
        "connection_manager",
        "current_address",
        "daemon_mode",
        "devices",
        "group_address_dpt",
        "knxip_interface",
        "management",
        "multicast_group",
        "multicast_port",
        "rate_limit",
        "sigint_received",
        "started",
        "state_updater",
        "task_registry",
        "telegram_queue",
        "telegrams",
        "version",
    )

    def __init__(
        self,
        address_format: GroupAddressType = GroupAddressType.LONG,
        telegram_received_cb: TelegramCallbackType | None = None,
        device_updated_cb: DeviceCallbackType[Device] | None = None,
        connection_state_changed_cb: ConnectionChangeCallbackType | None = None,
        rate_limit: int = 0,
        multicast_group: str = DEFAULT_MCAST_GRP,
        multicast_port: int = DEFAULT_MCAST_PORT,
        log_directory: str | None = None,
        state_updater: TrackerOptionType = False,
        daemon_mode: bool = False,
        connection_config: ConnectionConfig | None = None,
    ) -> None:
        """Initialize XKNX class."""
        self.connection_manager = ConnectionManager()
        self.knxip_interface = knx_interface_factory(
            self, connection_config=connection_config or ConnectionConfig()
        )
        self.management = Management(self)
        self.telegrams: asyncio.Queue[Telegram | None] = asyncio.Queue()
        self.telegram_queue = TelegramQueue(self)
        self.cemi_handler = CEMIHandler(self)
        self.state_updater = StateUpdater(self, default_tracker_option=state_updater)
        self.task_registry = TaskRegistry(self)
        self.group_address_dpt = GroupAddressDPT()

        self.current_address = IndividualAddress(0)
        self.daemon_mode = daemon_mode
        self.multicast_group = multicast_group
        self.multicast_port = multicast_port
        self.rate_limit = rate_limit
        self.sigint_received = asyncio.Event()
        self.started = asyncio.Event()
        self.devices = Devices(started=self.started)
        self.version = VERSION

        GroupAddress.address_format = address_format  # for global string representation
        if log_directory is not None:
            self.setup_logging(log_directory)

        if telegram_received_cb is not None:
            self.telegram_queue.register_telegram_received_cb(telegram_received_cb)

        if device_updated_cb is not None:
            self.devices.register_device_updated_cb(device_updated_cb)

        if connection_state_changed_cb is not None:
            self.connection_manager.register_connection_state_changed_cb(
                connection_state_changed_cb
            )

    def __del__(self) -> None:
        """Destructor. Cleaning up if this was not done before."""
        if self.started.is_set():
            try:
                loop = asyncio.get_event_loop()
                loop.run_until_complete(self.stop())
            except RuntimeError as exp:
                logger.warning("Could not close loop, reason: %s", exp)

    async def __aenter__(self) -> XKNX:
        """Start XKNX from context manager."""
        await self.start()
        return self

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc: BaseException | None,
        traceback: TracebackType | None,
    ) -> None:
        """Stop XKNX from context manager."""
        await self.stop()

    async def start(self) -> None:
        """Start XKNX module. Connect to KNX/IP devices and start state updater."""
        if self.knxip_interface.connection_config.threaded:
            await self.connection_manager.register_loop()
        self.task_registry.start()
        logger.info(
            "XKNX v%s starting %s connection to KNX bus.",
            VERSION,
            self.knxip_interface.connection_config.connection_type.name.lower(),
        )
        await self.knxip_interface.start()
        await self.telegram_queue.start()
        self.state_updater.start()
        self.devices.async_start_device_tasks()
        self.started.set()
        if self.daemon_mode:
            await self.loop_until_sigint()

    async def join(self) -> None:
        """Wait until all telegrams were processed."""
        await self.telegrams.join()

    async def stop(self) -> None:
        """Stop XKNX module."""
        self.devices.async_remove_device_tasks()
        self.task_registry.stop()
        self.state_updater.stop()
        await self.join()
        await self.telegram_queue.stop()
        await self.knxip_interface.stop()
        self.started.clear()

    async def loop_until_sigint(self) -> None:
        """Loop until Crtl-C was pressed."""

        def sigint_handler() -> None:
            """End loop."""
            self.sigint_received.set()

        if platform == "win32":
            logger.warning("Windows does not support signals")
        else:
            loop = asyncio.get_running_loop()
            loop.add_signal_handler(signal.SIGINT, sigint_handler)
        logger.warning("Press Ctrl+C to stop")
        await self.sigint_received.wait()

    @staticmethod
    def setup_logging(log_directory: str) -> None:
        """Configure logging to file."""
        log_path = Path(log_directory)
        if not log_path.is_dir():
            logger.warning("The provided log directory does not exist.")
            return

        _handler = TimedRotatingFileHandler(
            filename=log_path / "xknx.log",
            when="midnight",
            backupCount=7,
            encoding="utf-8",
        )
        _formatter = logging.Formatter(
            "%(asctime)s | %(name)s | %(levelname)s | %(message)s",
            datefmt="%Y-%m-%d %H:%M:%S",
        )
        _handler.setFormatter(_formatter)
        _handler.setLevel(logging.DEBUG)

        for log_namespace in [
            "xknx.cemi",
            "xknx.log",
            "xknx.knx",
            "xknx.raw_socket",
            "xknx.telegram",
            "xknx.state_updater",
        ]:
            _logger = logging.getLogger(log_namespace)
            _logger.addHandler(_handler)



"""Classes for handling KNX IP Secure."""

from .keyring import Keyring, load_keyring
from .util import bytes_xor, sha256_hash

__all__ = [
    "Keyring",
    "bytes_xor",
    "load_keyring",
    "sha256_hash",
]



"""Unit test for Weather objects."""

import datetime

import pytest

from xknx import XKNX
from xknx.devices import Weather
from xknx.devices.weather import WeatherCondition
from xknx.dpt import DPTArray, DPTBinary
from xknx.telegram import GroupAddress, Telegram
from xknx.telegram.apci import GroupValueWrite


class TestWeather:
    """Test class for Weather objects."""

    async def test_temperature(self) -> None:
        """Test resolve state with temperature."""
        xknx = XKNX()
        weather = Weather(name="weather", xknx=xknx, group_address_temperature="1/3/4")

        weather.process(
            Telegram(
                destination_address=GroupAddress("1/3/4"),
                payload=GroupValueWrite(value=DPTArray((0x19, 0xA))),
            )
        )

        assert weather.has_group_address(GroupAddress("1/3/4"))
        assert weather.temperature == 21.28
        assert weather._temperature.unit_of_measurement == "°C"
        assert weather._temperature.ha_device_class == "temperature"

    async def test_brightness(self) -> None:
        """Test resolve state for brightness east, west and south."""
        xknx = XKNX()
        weather: Weather = Weather(
            name="weather",
            xknx=xknx,
            group_address_brightness_east="1/3/5",
            group_address_brightness_south="1/3/6",
            group_address_brightness_west="1/3/7",
            group_address_brightness_north="1/3/8",
            group_address_temperature="1/4/4",
        )

        weather.process(
            Telegram(
                destination_address=GroupAddress("1/3/5"),
                payload=GroupValueWrite(value=DPTArray((0x7C, 0x5E))),
            )
        )

        weather.process(
            Telegram(
                destination_address=GroupAddress("1/3/7"),
                payload=GroupValueWrite(value=DPTArray((0x7C, 0x5C))),
            )
        )

        weather.process(
            Telegram(
                destination_address=GroupAddress("1/3/6"),
                payload=GroupValueWrite(value=DPTArray((0x7C, 0x5A))),
            )
        )

        weather.process(
            Telegram(
                destination_address=GroupAddress("1/3/8"),
                payload=GroupValueWrite(value=DPTArray((0x7C, 0x5A))),
            )
        )

        assert weather.brightness_east == 366346.24
        assert weather._brightness_east.unit_of_measurement == "lx"
        assert weather._brightness_east.ha_device_class == "illuminance"

        assert weather.brightness_west == 365690.88
        assert weather._brightness_west.unit_of_measurement == "lx"
        assert weather._brightness_west.ha_device_class == "illuminance"

        assert weather.brightness_south == 365035.52
        assert weather._brightness_south.unit_of_measurement == "lx"
        assert weather._brightness_south.ha_device_class == "illuminance"

        assert weather.brightness_north == 365035.52
        assert weather._brightness_north.unit_of_measurement == "lx"
        assert weather._brightness_north.ha_device_class == "illuminance"

    @pytest.mark.parametrize(
        ("value", "payload"),
        [
            (98631.68, DPTArray((0x6C, 0xB4))),  # 2byte float
            (98631.68, DPTArray((0x47, 0xC0, 0xA3, 0xD7))),  # 4byte float
        ],
    )
    async def test_pressure(self, value: float, payload: DPTArray) -> None:
        """Test air pressure telegram."""
        xknx = XKNX()
        weather = Weather(name="weather", xknx=xknx, group_address_air_pressure="1/3/4")

        weather.process(
            Telegram(
                destination_address=GroupAddress("1/3/4"),
                payload=GroupValueWrite(value=payload),
            )
        )

        assert weather.air_pressure == value
        assert weather._air_pressure.unit_of_measurement == "Pa"
        assert weather._air_pressure.ha_device_class == "pressure"

    async def test_humidity(self) -> None:
        """Test humidity telegram."""
        xknx = XKNX()
        weather = Weather(name="weather", xknx=xknx, group_address_humidity="1/2/4")

        weather.process(
            Telegram(
                destination_address=GroupAddress("1/2/4"),
                payload=GroupValueWrite(value=DPTArray((0x15, 0x73))),
            )
        )

        assert weather.humidity == 55.8
        assert weather._humidity.unit_of_measurement == "%"
        assert weather._humidity.ha_device_class == "humidity"

    async def test_wind_speed(self) -> None:
        """Test wind speed received."""
        xknx = XKNX()
        weather: Weather = Weather(
            name="weather", xknx=xknx, group_address_wind_speed="1/3/8"
        )

        weather.process(
            Telegram(
                destination_address=GroupAddress("1/3/8"),
                payload=GroupValueWrite(value=DPTArray((0x7D, 0x98))),
            )
        )

        assert weather.wind_speed == 469237.76
        assert weather._wind_speed.unit_of_measurement == "m/s"
        assert weather._wind_speed.ha_device_class == "wind_speed"

    async def test_wind_bearing(self) -> None:
        """Test wind bearing received."""
        xknx = XKNX()
        weather: Weather = Weather(
            name="weather", xknx=xknx, group_address_wind_bearing="1/3/8"
        )

        weather.process(
            Telegram(
                destination_address=GroupAddress("1/3/8"),
                payload=GroupValueWrite(value=DPTArray((0xBF,))),
            )
        )

        assert weather.wind_bearing == 270
        assert weather._wind_bearing.unit_of_measurement == "°"
        assert weather._wind_bearing.ha_device_class is None

    def test_state_lightning(self) -> None:
        """Test current_state returns lightning if wind alarm and rain alarm are true."""
        xknx = XKNX()
        weather: Weather = Weather(
            name="weather",
            xknx=xknx,
            group_address_rain_alarm="1/3/8",
            group_address_wind_alarm="1/3/9",
        )

        weather._rain_alarm.value = True
        weather._wind_alarm.value = True

        assert weather.ha_current_state() == WeatherCondition.LIGHTNING_RAINY

    def test_state_snowy_rainy(self) -> None:
        """Test snow rain if frost alarm and rain alarm are true."""
        xknx = XKNX()
        weather: Weather = Weather(
            name="weather",
            xknx=xknx,
            group_address_rain_alarm="1/3/8",
            group_address_frost_alarm="1/3/10",
        )

        weather._rain_alarm.value = True
        weather._frost_alarm.value = True

        assert weather.ha_current_state() == WeatherCondition.SNOWY_RAINY

    def test_wind_alarm(self) -> None:
        """Test basic state mapping."""
        xknx = XKNX()
        weather: Weather = Weather(
            name="weather",
            xknx=xknx,
            group_address_rain_alarm="1/3/8",
            group_address_wind_alarm="1/3/9",
            group_address_frost_alarm="1/3/10",
        )

        weather._wind_alarm.value = True

        assert weather.ha_current_state() == WeatherCondition.WINDY

    def test_rain_alarm(self) -> None:
        """Test basic state mapping."""
        xknx = XKNX()
        weather: Weather = Weather(
            name="weather",
            xknx=xknx,
            group_address_rain_alarm="1/3/8",
            group_address_wind_alarm="1/3/9",
            group_address_frost_alarm="1/3/10",
        )

        weather._rain_alarm.value = True

        assert weather.ha_current_state() == WeatherCondition.RAINY

    async def test_cloudy_summer(self) -> None:
        """Test cloudy summer if illuminance matches defined interval."""
        xknx = XKNX()
        weather: Weather = Weather(
            name="weather",
            xknx=xknx,
            group_address_brightness_east="1/3/5",
            group_address_brightness_south="1/3/6",
        )

        weather.process(
            Telegram(
                destination_address=GroupAddress("1/3/6"),
                payload=GroupValueWrite(value=DPTArray((0x46, 0x45))),
            )
        )

        summer_date = datetime.datetime(2020, 10, 5, 18, 00)

        assert (
            weather.ha_current_state(current_date=summer_date)
            == WeatherCondition.CLOUDY
        )

    async def test_sunny_summer(self) -> None:
        """Test returns sunny condition if illuminance is in defined interval."""
        xknx = XKNX()
        weather: Weather = Weather(
            name="weather",
            xknx=xknx,
            group_address_brightness_east="1/3/5",
            group_address_brightness_south="1/3/6",
            group_address_brightness_west="1/3/7",
        )

        weather.process(
            Telegram(
                destination_address=GroupAddress("1/3/6"),
                payload=GroupValueWrite(value=DPTArray((0x7C, 0x5C))),
            )
        )

        summer_date = datetime.datetime(2020, 10, 5, 18, 00)

        assert (
            weather.ha_current_state(current_date=summer_date) == WeatherCondition.SUNNY
        )

    async def test_sunny_winter(self) -> None:
        """Test sunny winter if illuminance matches defined interval."""
        xknx = XKNX()
        weather: Weather = Weather(
            name="weather",
            xknx=xknx,
            group_address_brightness_south="1/3/6",
            group_address_brightness_west="1/3/7",
        )

        weather.process(
            Telegram(
                destination_address=GroupAddress("1/3/6"),
                payload=GroupValueWrite(value=DPTArray((0x7C, 0x5C))),
            )
        )

        winter_date = datetime.datetime(2020, 12, 5, 18, 00)

        assert (
            weather.ha_current_state(current_date=winter_date) == WeatherCondition.SUNNY
        )

    async def test_cloudy_winter(self) -> None:
        """Test cloudy winter if illuminance matches defined interval."""
        xknx = XKNX()
        weather: Weather = Weather(
            name="weather",
            xknx=xknx,
            group_address_brightness_east="1/3/5",
            group_address_brightness_south="1/3/6",
            group_address_brightness_west="1/3/7",
        )

        weather.process(
            Telegram(
                destination_address=GroupAddress("1/3/6"),
                payload=GroupValueWrite(value=DPTArray((0x46, 0x45))),
            )
        )

        winter_date = datetime.datetime(2020, 12, 31, 18, 00)

        assert (
            weather.ha_current_state(current_date=winter_date)
            == WeatherCondition.CLOUDY
        )

    @pytest.mark.parametrize("invert_day_night", [False, True])
    async def test_day_night(self, invert_day_night: bool) -> None:
        """Test day night mapping."""
        xknx = XKNX()
        weather: Weather = Weather(
            name="weather",
            xknx=xknx,
            group_address_day_night="1/3/20",
            invert_day_night=invert_day_night,
        )

        weather.process(
            Telegram(
                destination_address=GroupAddress("1/3/20"),
                payload=GroupValueWrite(value=DPTBinary(0)),
            )
        )
        if invert_day_night:
            assert weather.ha_current_state() == WeatherCondition.EXCEPTIONAL
            assert weather.day_night is True
        else:
            assert weather.ha_current_state() == WeatherCondition.CLEAR_NIGHT
            assert weather.day_night is False

    def test_weather_default(self) -> None:
        """Test default state mapping."""
        xknx = XKNX()
        weather: Weather = Weather(name="weather", xknx=xknx)

        assert weather.ha_current_state() == WeatherCondition.EXCEPTIONAL

    #
    # GENERATOR _iter_remote_values
    #
    def test_iter_remote_values(self) -> None:
        """Test sensor has group address."""
        xknx = XKNX()
        weather = Weather(
            name="weather",
            xknx=xknx,
            group_address_temperature="1/3/4",
            group_address_rain_alarm="1/4/5",
            group_address_brightness_south="7/7/0",
        )
        assert weather.has_group_address(GroupAddress("1/3/4"))
        assert weather.has_group_address(GroupAddress("7/7/0"))
        assert weather.has_group_address(GroupAddress("1/4/5"))
        assert not weather.has_group_address(GroupAddress("1/2/4"))

    #
    # HAS GROUP ADDRESS
    #
    def test_has_group_address(self) -> None:
        """Test sensor has group address."""
        xknx = XKNX()
        weather = Weather(name="weather", xknx=xknx, group_address_temperature="1/3/4")
        assert weather.has_group_address(GroupAddress("1/3/4"))
        assert not weather.has_group_address(GroupAddress("1/2/4"))



"""
Module for managing a light via KNX.

It provides functionality for

* switching light 'on' and 'off'.
* setting the brightness.
* setting the color.
* setting the relative color temperature (tunable white).
* setting the absolute color temperature.
* reading the current state from KNX bus.
"""

from __future__ import annotations

from collections.abc import Iterator
from enum import Enum
from itertools import chain
import logging
from typing import TYPE_CHECKING, Any, cast

from xknx.core import Task
from xknx.dpt import RGBColor, RGBWColor, XYYColor
from xknx.remote_value import (
    GroupAddressesType,
    RemoteValue,
    RemoteValueColorRGB,
    RemoteValueColorRGBW,
    RemoteValueColorXYY,
    RemoteValueNumeric,
    RemoteValueScaling,
    RemoteValueSwitch,
)
from xknx.remote_value.remote_value import RVCallbackType

from .device import Device, DeviceCallbackType

if TYPE_CHECKING:
    from xknx.telegram import Telegram
    from xknx.xknx import XKNX

logger = logging.getLogger("xknx.log")


class ColorTemperatureType(Enum):
    """DPT used for absolute color temperature."""

    UINT_2_BYTE = "color_temperature"  # DPTColorTemperature 7.600
    FLOAT_2_BYTE = "2byte_float"  # DPT2ByteFloat generic 9


class _SwitchAndBrightness:
    def __init__(
        self,
        xknx: XKNX,
        name: str,
        feature_name: str,
        group_address_switch: GroupAddressesType = None,
        group_address_switch_state: GroupAddressesType = None,
        group_address_brightness: GroupAddressesType = None,
        group_address_brightness_state: GroupAddressesType = None,
        sync_state: bool | int | float | str = True,
        after_update_cb: RVCallbackType[bool | int] | None = None,
    ) -> None:
        self.switch = RemoteValueSwitch(
            xknx,
            group_address_switch,
            group_address_switch_state,
            sync_state=sync_state,
            device_name=name,
            feature_name=f"{feature_name}_state",
            after_update_cb=after_update_cb,
        )
        self.brightness = RemoteValueScaling(
            xknx,
            group_address_brightness,
            group_address_brightness_state,
            sync_state=sync_state,
            device_name=name,
            feature_name=f"{feature_name}_brightness",
            after_update_cb=after_update_cb,
            range_from=0,
            range_to=255,
        )

    @property
    def is_on(self) -> bool | None:
        """Return if light is on."""
        if self.switch.initialized:
            return self.switch.value
        if self.brightness.initialized and self.brightness.value is not None:
            return bool(self.brightness.value)
        return None

    def set_on(self) -> None:
        """Switch light on."""
        if self.switch.writable:
            self.switch.on()
            return
        if self.brightness.writable:
            self.brightness.set(self.brightness.range_to)

    def set_off(self) -> None:
        """Switch light off."""
        if self.switch.writable:
            self.switch.off()
            return
        if self.brightness.writable:
            self.brightness.set(0)

    def __eq__(self, other: object) -> bool:
        """Compare for equality."""
        return self.__dict__ == other.__dict__


class Light(Device):
    """Class for managing a light."""

    DEBOUNCE_TIMEOUT = 0.2
    DEFAULT_MIN_KELVIN = 2700  # 370 mireds
    DEFAULT_MAX_KELVIN = 6000  # 166 mireds

    def __init__(
        self,
        xknx: XKNX,
        name: str,
        group_address_switch: GroupAddressesType = None,
        group_address_switch_state: GroupAddressesType = None,
        group_address_brightness: GroupAddressesType = None,
        group_address_brightness_state: GroupAddressesType = None,
        group_address_color: GroupAddressesType = None,
        group_address_color_state: GroupAddressesType = None,
        group_address_rgbw: GroupAddressesType = None,
        group_address_rgbw_state: GroupAddressesType = None,
        group_address_hue: GroupAddressesType = None,
        group_address_hue_state: GroupAddressesType = None,
        group_address_saturation: GroupAddressesType = None,
        group_address_saturation_state: GroupAddressesType = None,
        group_address_xyy_color: GroupAddressesType = None,
        group_address_xyy_color_state: GroupAddressesType = None,
        group_address_tunable_white: GroupAddressesType = None,
        group_address_tunable_white_state: GroupAddressesType = None,
        group_address_color_temperature: GroupAddressesType = None,
        group_address_color_temperature_state: GroupAddressesType = None,
        group_address_switch_red: GroupAddressesType = None,
        group_address_switch_red_state: GroupAddressesType = None,
        group_address_brightness_red: GroupAddressesType = None,
        group_address_brightness_red_state: GroupAddressesType = None,
        group_address_switch_green: GroupAddressesType = None,
        group_address_switch_green_state: GroupAddressesType = None,
        group_address_brightness_green: GroupAddressesType = None,
        group_address_brightness_green_state: GroupAddressesType = None,
        group_address_switch_blue: GroupAddressesType = None,
        group_address_switch_blue_state: GroupAddressesType = None,
        group_address_brightness_blue: GroupAddressesType = None,
        group_address_brightness_blue_state: GroupAddressesType = None,
        group_address_switch_white: GroupAddressesType = None,
        group_address_switch_white_state: GroupAddressesType = None,
        group_address_brightness_white: GroupAddressesType = None,
        group_address_brightness_white_state: GroupAddressesType = None,
        color_temperature_type: ColorTemperatureType = ColorTemperatureType.UINT_2_BYTE,
        sync_state: bool | int | float | str = True,
        min_kelvin: int | None = None,
        max_kelvin: int | None = None,
        device_updated_cb: DeviceCallbackType[Light] | None = None,
    ) -> None:
        """Initialize Light class."""
        super().__init__(xknx, name, device_updated_cb)

        self.switch = RemoteValueSwitch(
            xknx,
            group_address_switch,
            group_address_switch_state,
            sync_state=sync_state,
            device_name=self.name,
            feature_name="State",
            after_update_cb=self.after_update,
        )

        self.brightness = RemoteValueScaling(
            xknx,
            group_address_brightness,
            group_address_brightness_state,
            sync_state=sync_state,
            device_name=self.name,
            feature_name="Brightness",
            after_update_cb=self.after_update,
            range_from=0,
            range_to=255,
        )

        self.color = RemoteValueColorRGB(
            xknx,
            group_address_color,
            group_address_color_state,
            sync_state=sync_state,
            device_name=self.name,
            feature_name="Color RGB",
            after_update_cb=self.after_update,
        )

        self.rgbw = RemoteValueColorRGBW(
            xknx,
            group_address_rgbw,
            group_address_rgbw_state,
            sync_state=sync_state,
            device_name=self.name,
            feature_name="Color RGBW",
            after_update_cb=self.after_update,
        )

        self.hue = RemoteValueNumeric(
            xknx,
            group_address_hue,
            group_address_hue_state,
            sync_state=sync_state,
            value_type="angle",
            device_name=self.name,
            feature_name="Hue",
            after_update_cb=self.after_update,
        )

        self.saturation = RemoteValueNumeric(
            xknx,
            group_address_saturation,
            group_address_saturation_state,
            sync_state=sync_state,
            value_type="percent",
            device_name=self.name,
            feature_name="Saturation",
            after_update_cb=self.after_update,
        )

        self._xyy_color_valid: XYYColor | None = None
        self.xyy_color = RemoteValueColorXYY(
            xknx,
            group_address_xyy_color,
            group_address_xyy_color_state,
            sync_state=sync_state,
            device_name=self.name,
            feature_name="Color XYY",
            after_update_cb=self._xyy_color_from_rv,
        )

        self.tunable_white = RemoteValueScaling(
            xknx,
            group_address_tunable_white,
            group_address_tunable_white_state,
            sync_state=sync_state,
            device_name=self.name,
            feature_name="Tunable white",
            after_update_cb=self.after_update,
            range_from=0,
            range_to=255,
        )

        self.color_temperature = RemoteValueNumeric(
            xknx,
            group_address_color_temperature,
            group_address_color_temperature_state,
            sync_state=sync_state,
            value_type=color_temperature_type.value,
            device_name=self.name,
            feature_name="Color temperature",
            after_update_cb=self.after_update,
        )

        self.red = _SwitchAndBrightness(
            xknx,
            self.name,
            "red",
            group_address_switch_red,
            group_address_switch_red_state,
            group_address_brightness_red,
            group_address_brightness_red_state,
            sync_state=sync_state,
            after_update_cb=self._individual_color_callback_debounce,
        )

        self.green = _SwitchAndBrightness(
            xknx,
            self.name,
            "green",
            group_address_switch_green,
            group_address_switch_green_state,
            group_address_brightness_green,
            group_address_brightness_green_state,
            sync_state=sync_state,
            after_update_cb=self._individual_color_callback_debounce,
        )

        self.blue = _SwitchAndBrightness(
            xknx,
            self.name,
            "blue",
            group_address_switch_blue,
            group_address_switch_blue_state,
            group_address_brightness_blue,
            group_address_brightness_blue_state,
            sync_state=sync_state,
            after_update_cb=self._individual_color_callback_debounce,
        )

        self.white = _SwitchAndBrightness(
            xknx,
            self.name,
            "white",
            group_address_switch_white,
            group_address_switch_white_state,
            group_address_brightness_white,
            group_address_brightness_white_state,
            sync_state=sync_state,
            after_update_cb=self._individual_color_callback_debounce,
        )

        self.min_kelvin = min_kelvin
        self.max_kelvin = max_kelvin
        self._individual_color_debounce_task = Task(
            name=f"{id(self)}_individual_color_debounce",
            target=self._debouncer_finished,
            wait_before_start=self.DEBOUNCE_TIMEOUT,
        )
        self._individual_color_debounce_telegram_counter = (
            self._initial_individual_color_debounce_telegrams()
        )

    def _iter_remote_values(self) -> Iterator[RemoteValue[Any]]:
        """Iterate the devices RemoteValue classes."""
        return chain(
            self._iter_instant_remote_values(),
            self._iter_debounce_remote_values(),
        )

    def _iter_instant_remote_values(self) -> Iterator[RemoteValue[Any]]:
        """Iterate the devices RemoteValue classes calling after_update_cb immediately."""
        yield self.switch
        yield self.brightness
        yield self.color
        yield self.rgbw
        yield self.hue
        yield self.saturation
        yield self.xyy_color
        yield self.tunable_white
        yield self.color_temperature

    def _iter_debounce_remote_values(self) -> Iterator[RemoteValue[Any]]:
        """Iterate the devices RemoteValue classes debouncing after_update_cb."""
        for color in self._iter_individual_colors():
            yield color.switch
            yield color.brightness

    def _iter_individual_colors(self) -> Iterator[_SwitchAndBrightness]:
        """Iterate the devices individual colors."""
        yield from (self.red, self.green, self.blue, self.white)

    def async_remove_tasks(self) -> None:
        """Remove async tasks of device."""
        self.xknx.task_registry.remove_task(self._individual_color_debounce_task)

    def _initial_individual_color_debounce_telegrams(self) -> int:
        """Reset individual color debounce telegram counter."""
        return sum(
            (
                self.red.switch.initialized or self.red.brightness.initialized,
                self.green.switch.initialized or self.green.brightness.initialized,
                self.blue.switch.initialized or self.blue.brightness.initialized,
                self.white.switch.initialized or self.white.brightness.initialized,
            )
        )

    def _debouncer_finished(self) -> None:
        """Reset debouncer after all telegrams were processed."""
        self._individual_color_debounce_telegram_counter = (
            self._initial_individual_color_debounce_telegrams()
        )
        self.after_update()

    def _individual_color_callback_debounce(self, *args: Any) -> None:
        """Run callback after all individual colors were updated or timeout passed."""
        self._individual_color_debounce_telegram_counter -= 1
        if self._individual_color_debounce_telegram_counter > 0:
            # task registry cancels existing task
            self.xknx.task_registry.start_task(self._individual_color_debounce_task)
            return
        self._individual_color_debounce_task.cancel()
        self._debouncer_finished()

    @property
    def supports_brightness(self) -> bool:
        """Return if light supports brightness."""
        return self.brightness.initialized

    @property
    def supports_color(self) -> bool:
        """Return if light supports color."""
        return self.color.initialized or all(
            c.brightness.initialized for c in (self.red, self.green, self.blue)
        )

    @property
    def supports_rgbw(self) -> bool:
        """Return if light supports RGBW."""
        return self.rgbw.initialized or all(
            c.brightness.initialized for c in self._iter_individual_colors()
        )

    @property
    def supports_hs_color(self) -> bool:
        """Return if light supports HS-color."""
        return self.hue.initialized and self.saturation.initialized

    @property
    def supports_xyy_color(self) -> bool:
        """Return if light supports xyY-color."""
        return self.xyy_color.initialized

    @property
    def supports_tunable_white(self) -> bool:
        """Return if light supports tunable white / relative color temperature."""
        return self.tunable_white.initialized

    @property
    def supports_color_temperature(self) -> bool:
        """Return if light supports absolute color temperature."""
        return self.color_temperature.initialized

    @property
    def state(self) -> bool | None:
        """Return the current switch state of the device."""
        if self.switch.value is not None:
            return self.switch.value
        if any(c.is_on is not None for c in self._iter_individual_colors()):
            return any(c.is_on for c in self._iter_individual_colors())
        return None

    async def set_on(self) -> None:
        """Switch light on."""
        if self.switch.writable:
            self.switch.on()
            return
        for color in self._iter_individual_colors():
            color.set_on()

    async def set_off(self) -> None:
        """Switch light off."""
        if self.switch.writable:
            self.switch.off()
            return
        for color in self._iter_individual_colors():
            color.set_off()

    @property
    def current_brightness(self) -> int | None:
        """Return current brightness of light between 0..255."""
        return self.brightness.value

    async def set_brightness(self, brightness: int) -> None:
        """Set brightness of light."""
        if not self.supports_brightness:
            logger.warning("Dimming not supported for device %s", self.get_name())
            return
        self.brightness.set(brightness)

    @property
    def current_color(self) -> tuple[tuple[int, int, int] | None, int | None]:
        """
        Return current color of light.

        If the device supports RGBW, get the current RGB+White values instead.
        """
        if self.supports_rgbw and self.rgbw.initialized:
            if not self.rgbw.value:
                return None, None
            if (
                self.rgbw.value.red is not None
                and self.rgbw.value.green is not None
                and self.rgbw.value.blue is not None
            ):
                return (
                    (self.rgbw.value.red, self.rgbw.value.green, self.rgbw.value.blue),
                    self.rgbw.value.white,
                )
            return None, self.rgbw.value.white
        if self.color.initialized:
            if self.color.value is None:
                return None, None
            return (
                self.color.value.red,
                self.color.value.green,
                self.color.value.blue,
            ), None
        # individual RGB addresses - white will return None when it is not initialized
        colors = (
            self.red.brightness.value,
            self.green.brightness.value,
            self.blue.brightness.value,
        )
        if None in colors:
            return None, self.white.brightness.value
        return cast(tuple[int, int, int], colors), self.white.brightness.value

    async def set_color(
        self, color: tuple[int, int, int], white: int | None = None
    ) -> None:
        """
        Set color of a light device.

        If also the white value is given and the device supports RGBW,
        set all four values.
        """
        if white is not None:
            if self.supports_rgbw:
                if self.rgbw.initialized:
                    self.rgbw.set(RGBWColor(*color, white))
                    return
                if all(
                    c.brightness.initialized for c in self._iter_individual_colors()
                ):
                    self.red.brightness.set(color[0])
                    self.green.brightness.set(color[1])
                    self.blue.brightness.set(color[2])
                    self.white.brightness.set(white)
                    return
            logger.warning("RGBW not supported for device %s", self.get_name())
        else:
            if self.supports_color:
                if self.color.initialized:
                    self.color.set(RGBColor(*color))
                    return
                if all(
                    c.brightness.initialized for c in (self.red, self.green, self.blue)
                ):
                    self.red.brightness.set(color[0])
                    self.green.brightness.set(color[1])
                    self.blue.brightness.set(color[2])
                    return
            logger.warning("Colors not supported for device %s", self.get_name())

    @property
    def current_hs_color(self) -> tuple[float, float] | None:
        """
        Return current HS-color of the light.

        Hue is scaled 0-360 (265 possible values from KNX)
        Sat is scaled 0-100
        """
        if (hue := self.hue.value) is not None and (
            (saturation := self.saturation.value) is not None
        ):
            return (hue, saturation)
        return None

    async def set_hs_color(self, hs_color: tuple[float, float]) -> None:
        """Set HS-color of the light."""
        if not self.supports_hs_color:
            logger.warning("HS-color not supported for device %s", self.get_name())
            return
        value_sent = False
        if (hue := hs_color[0]) != self.hue.value:
            self.hue.set(hue)
            value_sent = True
        if (saturation := hs_color[1]) != self.saturation.value:
            self.saturation.set(saturation)
            value_sent = True
        if not value_sent:
            # at least one value shall be sent to enable turn-on by hs_color
            self.hue.set(hue)
            self.saturation.set(saturation)

    def _xyy_color_from_rv(self, xyy_color: XYYColor) -> None:
        """Update the current xyY-color from RemoteValue (Callback)."""
        if self._xyy_color_valid is not None:
            self._xyy_color_valid = self._xyy_color_valid | xyy_color
        else:
            self._xyy_color_valid = xyy_color
        self.after_update()

    @property
    def current_xyy_color(self) -> XYYColor | None:
        """Return current xyY-color of the light."""
        return self._xyy_color_valid

    async def set_xyy_color(self, xyy: XYYColor) -> None:
        """Set xyY-color of the light."""
        if not self.supports_xyy_color:
            logger.warning("XYY-color not supported for device %s", self.get_name())
            return
        self.xyy_color.set(xyy)

    @property
    def current_tunable_white(self) -> int | None:
        """Return current relative color temperature of light."""
        return self.tunable_white.value

    async def set_tunable_white(self, tunable_white: int) -> None:
        """Set relative color temperature of light."""
        if not self.supports_tunable_white:
            logger.warning("Tunable white not supported for device %s", self.get_name())
            return
        self.tunable_white.set(tunable_white)

    @property
    def current_color_temperature(self) -> int | float | None:
        """Return current absolute color temperature of light."""
        return self.color_temperature.value

    async def set_color_temperature(self, color_temperature: int) -> None:
        """Set absolute color temperature of light."""
        if not self.supports_color_temperature:
            logger.warning(
                "Absolute Color Temperature not supported for device %s",
                self.get_name(),
            )
            return
        self.color_temperature.set(color_temperature)

    def process_group_write(self, telegram: Telegram) -> None:
        """Process incoming and outgoing GROUP WRITE telegram."""
        for remote_value in self._iter_instant_remote_values():
            remote_value.process(telegram)
        for remote_value in self._iter_debounce_remote_values():
            remote_value.process(telegram, always_callback=True)

    def __str__(self) -> str:
        """Return object as readable string."""
        str_brightness = (
            f" brightness={self.brightness.group_addr_str()}"
            if self.supports_brightness
            else ""
        )

        str_color = (
            f" color={self.color.group_addr_str()}" if self.supports_color else ""
        )

        str_rgbw = f" rgbw={self.rgbw.group_addr_str()}" if self.supports_rgbw else ""

        str_hue = (
            f" brightness={self.hue.group_addr_str()}" if self.hue.initialized else ""
        )

        str_saturation = (
            f" brightness={self.saturation.group_addr_str()}"
            if self.saturation.initialized
            else ""
        )

        str_xyy_color = (
            f" xyy_color={self.xyy_color.group_addr_str()}"
            if self.supports_xyy_color
            else ""
        )
        str_tunable_white = (
            f" tunable_white={self.tunable_white.group_addr_str()}"
            if self.supports_tunable_white
            else ""
        )

        str_color_temperature = (
            f" color_temperature={self.color_temperature.group_addr_str()}"
            if self.supports_color_temperature
            else ""
        )

        str_red_state = (
            f" red_state={self.red.switch.group_addr_str()}"
            if self.red.switch.initialized
            else ""
        )
        str_red_brightness = (
            f" red_brightness={self.red.brightness.group_addr_str()}"
            if self.red.brightness.initialized
            else ""
        )

        str_green_state = (
            f" green_state={self.green.switch.group_addr_str()}"
            if self.green.switch.initialized
            else ""
        )
        str_green_brightness = (
            f" green_brightness={self.green.brightness.group_addr_str()}"
            if self.green.brightness.initialized
            else ""
        )

        str_blue_state = (
            f" blue_state={self.blue.switch.group_addr_str()}"
            if self.blue.switch.initialized
            else ""
        )
        str_blue_brightness = (
            f" blue_brightness={self.blue.brightness.group_addr_str()}"
            if self.blue.brightness.initialized
            else ""
        )

        str_white_state = (
            f" white_state={self.white.switch.group_addr_str()}"
            if self.white.switch.initialized
            else ""
        )
        str_white_brightness = (
            f" white_brightness={self.white.brightness.group_addr_str()}"
            if self.white.brightness.initialized
            else ""
        )

        return (
            f'<Light name="{self.name}" '
            f"switch={self.switch.group_addr_str()}"
            f"{str_brightness}"
            f"{str_color}"
            f"{str_rgbw}"
            f"{str_hue}"
            f"{str_saturation}"
            f"{str_xyy_color}"
            f"{str_tunable_white}"
            f"{str_color_temperature}"
            f"{str_red_state}"
            f"{str_red_brightness}"
            f"{str_green_state}"
            f"{str_green_brightness}"
            f"{str_blue_state}"
            f"{str_blue_brightness}"
            f"{str_white_state}"
            f"{str_white_brightness}"
            " />"
        )



"""Unit test for KNX/IP gateway scanner."""

import asyncio
from typing import Any
from unittest.mock import Mock, create_autospec, patch

import pytest

from xknx import XKNX
from xknx.exceptions import XKNXException
from xknx.io import GatewayScanFilter, GatewayScanner
from xknx.io.gateway_scanner import GatewayDescriptor
from xknx.io.transport import UDPTransport
from xknx.knxip import (
    HPAI,
    DIBDeviceInformation,
    DIBServiceFamily,
    DIBSuppSVCFamilies,
    KNXIPFrame,
    SearchRequest,
    SearchRequestExtended,
    SearchResponse,
    SearchResponseExtended,
)
from xknx.telegram import IndividualAddress

from ..conftest import EventLoopClockAdvancer


class TestGatewayDescriptor:
    """Test GatewayDescriptor object."""

    @pytest.mark.parametrize(
        "raw,expected",
        [
            (
                # MDT SCN-IP000.03 IP Interface (IP-Secure enabled)
                bytes.fromhex(
                    "0610020c00a20801c0a800ba0e57360102001101000000837b40054500000000"
                    "cc1be080b80153434e2d49503030302e303320495020496e7465726661636520"
                    "6d6974200c02020203020402070209010808000000f000121003000000004005"
                    "45000000000001041404c0a800ba40054500c0a800010000000004000c051101"
                    "1102110311041105060603010401140700f01102000011030000110400001105"
                    "0000"
                ),
                {
                    "supports_routing": False,
                    "supports_tunnelling": True,
                    "supports_tunnelling_tcp": True,
                    "supports_secure": True,
                    "routing_requires_secure": False,
                    "tunnelling_requires_secure": True,
                },
            ),
            (
                # MDT SCN-IP100.02 IP Router (no IP-Secure support, CoreV1)
                bytes.fromhex(
                    "06100202004e08010a0102280e5736010200200000000083477f0124e000170c"
                    "cc1be08004c44d4454204b4e5820495020526f75746572000000000000000000"
                    "000000000a020201030104010501"
                ),
                {
                    "supports_routing": True,
                    "supports_tunnelling": True,
                    "supports_tunnelling_tcp": False,
                    "supports_secure": False,
                    "routing_requires_secure": None,
                    "tunnelling_requires_secure": None,
                },
            ),
            (
                # Gira X1 (no IP-Secure support but CoreV2)
                bytes.fromhex(
                    "0610020c006408010a0100290e57360102001001000000081171012600000000"
                    "000ab3290b134769726120583100000000000000000000000000000000000000"
                    "000000000c0202020302040207010901140700dc10fbfff810fcffff10fdffff"
                    "10feffff"
                ),
                {
                    "supports_routing": False,
                    "supports_tunnelling": True,
                    "supports_tunnelling_tcp": True,
                    "supports_secure": True,
                    "routing_requires_secure": None,
                    "tunnelling_requires_secure": None,
                },
            ),
            (
                # Gira IP-Router I12 (no IP-Secure support but CoreV2)
                bytes.fromhex(
                    "0610020c006608010a0100280e57360102001000000000082d40834de000170c"
                    "000ab3274a3247697261204b4e582f49502d526f757465720000000000000000"
                    "000000000e02020203020402050207010901140700dc10f1fffe10f2ffff10f3"
                    "ffff10f4ffff"
                ),
                {
                    "supports_routing": True,
                    "supports_tunnelling": True,
                    "supports_tunnelling_tcp": True,
                    "supports_secure": True,
                    "routing_requires_secure": None,
                    "tunnelling_requires_secure": None,
                },
            ),
            (
                # Jung IPR 300 SREG (IP-Secure disabled)
                bytes.fromhex(
                    "0610020c007408010a0101550e57360102004000000000a615000037e000170c"
                    "0022d10400374a756e67204b4e582049502d526f757465720000000000000000"
                    "000000000c0202020302040205020901240700f8400100054002000540030005"
                    "4004000540050005400600054007000540080005"
                ),
                {
                    "supports_routing": True,
                    "supports_tunnelling": True,
                    "supports_tunnelling_tcp": True,
                    "supports_secure": True,
                    "routing_requires_secure": None,
                    "tunnelling_requires_secure": None,
                },
            ),
            (
                # Jung IPR 300 SREG (IP-Secure enabled)
                bytes.fromhex(
                    "0610020c008408010a0101550e57360102004000000000a615000037e000170c"
                    "0022d10400374b4e582049502d526f7574657200000000000000000000000000"
                    "000000000c02020203020402050209010808000004b0091a0806030104010501"
                    "240700f840010005400200054003000540040005400500054006000540070005"
                    "40080005"
                ),
                {
                    "supports_routing": True,
                    "supports_tunnelling": True,
                    "supports_tunnelling_tcp": True,
                    "supports_secure": True,
                    "routing_requires_secure": True,
                    "tunnelling_requires_secure": True,
                },
            ),
        ],
    )
    def test_parser(self, raw: bytes, expected: dict[str, Any]) -> None:
        """Test parsing GatewayDescriptor objects from real-world responses."""
        response, _ = KNXIPFrame.from_knx(raw)
        assert isinstance(response.body, SearchResponse | SearchResponseExtended)

        descriptor = GatewayDescriptor(
            ip_addr=response.body.control_endpoint.ip_addr,
            port=response.body.control_endpoint.port,
        )
        descriptor.parse_dibs(response.body.dibs)

        assert descriptor.supports_routing is expected["supports_routing"]
        assert descriptor.supports_tunnelling is expected["supports_tunnelling"]
        assert descriptor.supports_tunnelling_tcp is expected["supports_tunnelling_tcp"]

        assert descriptor.supports_secure is expected["supports_secure"]

        assert descriptor.routing_requires_secure is expected["routing_requires_secure"]
        assert (
            descriptor.tunnelling_requires_secure
            is expected["tunnelling_requires_secure"]
        )


class TestGatewayScanner:
    """Test class for xknx/io/GatewayScanner objects."""

    gateway_desc_interface = GatewayDescriptor(
        name="KNX-Interface",
        ip_addr="10.1.1.11",
        port=3761,
        local_interface="en1",
        local_ip="10.1.1.100",
        supports_tunnelling=True,
        supports_routing=False,
    )
    gateway_desc_router = GatewayDescriptor(
        name="KNX-Router",
        ip_addr="10.1.1.12",
        port=3761,
        local_interface="en1",
        local_ip="10.1.1.100",
        supports_tunnelling=False,
        supports_routing=True,
    )
    gateway_desc_both = GatewayDescriptor(
        name="Gira KNX/IP-Router",
        ip_addr="192.168.42.10",
        port=3671,
        local_interface="en1",
        local_ip="192.168.42.50",
        supports_tunnelling=True,
        supports_tunnelling_tcp=True,
        supports_routing=True,
        individual_address=IndividualAddress("1.1.0"),
    )
    gateway_desc_both.tunnelling_requires_secure = False
    gateway_desc_neither = GatewayDescriptor(
        name="AC/S 1.1.1 Application Control",
        ip_addr="10.1.1.15",
        port=3671,
        local_interface="en1",
        local_ip="192.168.42.50",
        supports_tunnelling=False,
        supports_routing=False,
    )
    gateway_desc_secure_tunnel = GatewayDescriptor(
        name="KNX-Interface",
        ip_addr="10.1.1.11",
        port=3761,
        local_interface="en1",
        local_ip="10.1.1.111",
        supports_routing=False,
        supports_tunnelling=True,
        supports_tunnelling_tcp=True,
        supports_secure=True,
    )
    gateway_desc_secure_tunnel.tunnelling_requires_secure = True
    gateway_desc_secure_router = GatewayDescriptor(
        name="KNX-Secure-Router",
        ip_addr="10.1.1.12",
        port=3761,
        local_interface="en1",
        local_ip="10.1.1.100",
        supports_tunnelling=False,
        supports_routing=True,
    )
    gateway_desc_secure_router.routing_requires_secure = True

    def test_gateway_scan_filter_match(self) -> None:
        """Test match function of gateway filter."""
        filter_default = GatewayScanFilter()
        filter_tunnel = GatewayScanFilter(routing=False, secure_routing=False)
        filter_tcp_tunnel = GatewayScanFilter(
            tunnelling=False,
            tunnelling_tcp=True,
            secure_tunnelling=None,
            routing=False,
            secure_routing=False,
        )
        filter_secure_tunnel = GatewayScanFilter(
            tunnelling=False,
            tunnelling_tcp=False,
            secure_tunnelling=True,
            routing=False,
            secure_routing=False,
        )
        filter_router = GatewayScanFilter(
            tunnelling=False,
            tunnelling_tcp=False,
            secure_tunnelling=False,
            routing=True,
            secure_routing=False,
        )
        filter_name = GatewayScanFilter(name="KNX-Router")
        filter_secure_router = GatewayScanFilter(
            tunnelling=False,
            tunnelling_tcp=False,
            secure_tunnelling=False,
            routing=False,
            secure_routing=True,
        )

        assert filter_default.match(self.gateway_desc_interface)
        assert filter_default.match(self.gateway_desc_router)
        assert filter_default.match(self.gateway_desc_both)
        assert not filter_default.match(self.gateway_desc_neither)
        assert filter_default.match(self.gateway_desc_secure_tunnel)
        assert filter_default.match(self.gateway_desc_secure_router)

        assert filter_tunnel.match(self.gateway_desc_interface)
        assert not filter_tunnel.match(self.gateway_desc_router)
        assert filter_tunnel.match(self.gateway_desc_both)
        assert not filter_tunnel.match(self.gateway_desc_neither)
        assert filter_tunnel.match(self.gateway_desc_secure_tunnel)
        assert not filter_tunnel.match(self.gateway_desc_secure_router)

        assert not filter_tcp_tunnel.match(self.gateway_desc_interface)
        assert not filter_tcp_tunnel.match(self.gateway_desc_router)
        assert filter_tcp_tunnel.match(self.gateway_desc_both)
        assert not filter_tcp_tunnel.match(self.gateway_desc_neither)
        assert not filter_tcp_tunnel.match(self.gateway_desc_secure_tunnel)
        assert not filter_tcp_tunnel.match(self.gateway_desc_secure_router)

        assert not filter_secure_tunnel.match(self.gateway_desc_interface)
        assert not filter_secure_tunnel.match(self.gateway_desc_router)
        assert not filter_secure_tunnel.match(self.gateway_desc_both)
        assert not filter_secure_tunnel.match(self.gateway_desc_neither)
        assert filter_secure_tunnel.match(self.gateway_desc_secure_tunnel)
        assert not filter_secure_tunnel.match(self.gateway_desc_secure_router)

        assert not filter_router.match(self.gateway_desc_interface)
        assert filter_router.match(self.gateway_desc_router)
        assert filter_router.match(self.gateway_desc_both)
        assert not filter_router.match(self.gateway_desc_neither)
        assert not filter_router.match(self.gateway_desc_secure_tunnel)
        assert not filter_router.match(self.gateway_desc_secure_router)

        assert not filter_name.match(self.gateway_desc_interface)
        assert filter_name.match(self.gateway_desc_router)
        assert not filter_name.match(self.gateway_desc_both)
        assert not filter_name.match(self.gateway_desc_neither)
        assert not filter_name.match(self.gateway_desc_secure_tunnel)
        assert not filter_name.match(self.gateway_desc_secure_router)

        assert not filter_secure_router.match(self.gateway_desc_interface)
        assert not filter_secure_router.match(self.gateway_desc_router)
        assert not filter_secure_router.match(self.gateway_desc_both)
        assert not filter_secure_router.match(self.gateway_desc_neither)
        assert not filter_secure_router.match(self.gateway_desc_secure_tunnel)
        assert filter_secure_router.match(self.gateway_desc_secure_router)

    def test_search_response_reception(self) -> None:
        """Test function of gateway scanner."""
        xknx = XKNX()
        gateway_scanner = GatewayScanner(xknx)
        test_search_response = fake_router_search_response()
        udp_transport_mock = create_autospec(UDPTransport)
        udp_transport_mock.local_addr = ("192.168.42.50", 0)
        udp_transport_mock.getsockname.return_value = ("192.168.42.50", 0)

        assert not gateway_scanner.found_gateways
        gateway_scanner._response_rec_callback(
            test_search_response,
            HPAI("192.168.42.50", 0),
            udp_transport_mock,
            interface="en1",
        )
        assert len(gateway_scanner.found_gateways) == 1

        gateway_scanner._response_rec_callback(
            test_search_response,
            HPAI("192.168.42.230", 0),
            udp_transport_mock,
            interface="eth1",
        )
        assert len(gateway_scanner.found_gateways) == 1

        assert str(
            gateway_scanner.found_gateways[test_search_response.body.control_endpoint]
        ) == str(self.gateway_desc_both)

    @patch("xknx.io.gateway_scanner.UDPTransport.connect")
    @patch("xknx.io.gateway_scanner.UDPTransport.send")
    @patch(
        "xknx.io.gateway_scanner.UDPTransport.getsockname",
        return_value=("10.1.1.2", 56789),
    )
    async def test_scan_timeout(
        self,
        getsockname_mock: Mock,
        udp_transport_send_mock: Mock,
        udp_transport_connect_mock: Mock,
        time_travel: EventLoopClockAdvancer,
    ) -> None:
        """Test gateway scanner timeout."""
        xknx = XKNX()
        gateway_scanner = GatewayScanner(xknx)
        timed_out_scan_task = asyncio.create_task(gateway_scanner.scan())
        await time_travel(gateway_scanner.timeout_in_seconds)
        # Unsuccessful scan() returns empty list
        assert await timed_out_scan_task == []

    @patch("xknx.io.gateway_scanner.UDPTransport.connect")
    @patch("xknx.io.gateway_scanner.UDPTransport.send")
    async def test_async_scan_timeout(
        self,
        udp_transport_send_mock: Mock,
        udp_transport_connect_mock: Mock,
        time_travel: EventLoopClockAdvancer,
    ) -> None:
        """Test gateway scanner timeout for async generator."""

        async def test() -> bool:
            xknx = XKNX()
            async for _ in GatewayScanner(xknx).async_scan():
                break
            else:
                return True

        # timeout
        with (
            patch(
                "xknx.io.util.get_default_local_ip",
                return_value="10.1.1.2",
            ),
            patch(
                "xknx.io.gateway_scanner.UDPTransport.getsockname",
                return_value=("10.1.1.2", 56789),
            ),
        ):
            timed_out_scan_task = asyncio.create_task(test())
            await time_travel(3)
            assert await timed_out_scan_task
        # no matching interface found
        with patch(
            "xknx.io.util.get_default_local_ip",
            return_value=None,
        ):
            timed_out_scan_task = asyncio.create_task(test())
            await time_travel(3)
            with pytest.raises(XKNXException):
                await timed_out_scan_task

    @patch("xknx.io.gateway_scanner.UDPTransport.connect")
    @patch("xknx.io.gateway_scanner.UDPTransport.send")
    async def test_async_scan_exit(
        self,
        udp_transport_send_mock: Mock,
        udp_transport_connect_mock: Mock,
        time_travel: EventLoopClockAdvancer,
    ) -> None:
        """Test gateway scanner timeout for async generator."""
        xknx = XKNX()
        test_search_response = fake_router_search_response()
        udp_transport_mock = Mock()
        udp_transport_mock.local_addr = ("10.1.1.2", 56789)

        gateway_scanner = GatewayScanner(xknx, local_ip="10.1.1.2")

        async def test() -> bool:
            async for gateway in gateway_scanner.async_scan():
                assert isinstance(gateway, GatewayDescriptor)
                return True
            return False

        with (
            patch(
                "xknx.io.gateway_scanner.UDPTransport.getsockname",
                return_value=("10.1.1.2", 56789),
            ),
            patch(
                "xknx.io.gateway_scanner.UDPTransport.register_callback"
            ) as register_callback_mock,
        ):
            scan_task = asyncio.create_task(test())
            await time_travel(0)
            _fished_response_rec_callback = register_callback_mock.call_args.args[0]
            _fished_response_rec_callback(
                test_search_response,
                HPAI("192.168.42.50", 0),
                udp_transport_mock,
            )
            assert await scan_task
            await time_travel(0)  # for task cleanup

    @patch("xknx.io.gateway_scanner.UDPTransport.connect")
    @patch("xknx.io.gateway_scanner.UDPTransport.send")
    async def test_send_search_requests(
        self,
        udp_transport_send_mock: Mock,
        udp_transport_connect_mock: Mock,
    ) -> None:
        """Test if both search requests are sent per interface."""
        xknx = XKNX()
        gateway_scanner = GatewayScanner(xknx, timeout_in_seconds=0)
        with (
            patch(
                "xknx.io.util.get_default_local_ip",
                return_value="10.1.1.2",
            ),
            patch(
                "xknx.io.util.get_local_interface_name",
                return_value="en_0123",
            ),
            patch(
                "xknx.io.gateway_scanner.UDPTransport.getsockname",
                return_value=("10.1.1.2", 56789),
            ),
        ):
            await gateway_scanner.scan()

        assert udp_transport_connect_mock.call_count == 1
        assert udp_transport_send_mock.call_count == 2
        frame_1 = udp_transport_send_mock.call_args_list[0].args[0]
        frame_2 = udp_transport_send_mock.call_args_list[1].args[0]
        assert isinstance(frame_1.body, SearchRequestExtended)
        assert isinstance(frame_2.body, SearchRequest)
        assert frame_1.body.discovery_endpoint == HPAI(ip_addr="10.1.1.2", port=56789)

    def test_gateway_scan_filter_compare(self) -> None:
        """Test GatewayScanFilter comparison."""
        assert GatewayScanFilter() == GatewayScanFilter()
        assert GatewayScanFilter() != GatewayScanFilter(tunnelling=False)


def fake_router_search_response() -> KNXIPFrame:
    """Return the KNXIPFrame of a KNX/IP Router with a SearchResponse body."""
    frame_body = SearchResponse()
    frame_body.control_endpoint = HPAI(ip_addr="192.168.42.10", port=3671)

    device_information = DIBDeviceInformation()
    device_information.name = "Gira KNX/IP-Router"
    device_information.serial_number = "11:22:33:44:55:66"
    device_information.individual_address = IndividualAddress("1.1.0")
    device_information.mac_address = "01:02:03:04:05:06"

    svc_families = DIBSuppSVCFamilies()
    svc_families.families.append(
        DIBSuppSVCFamilies.Family(name=DIBServiceFamily.CORE, version=1)
    )
    svc_families.families.append(
        DIBSuppSVCFamilies.Family(name=DIBServiceFamily.DEVICE_MANAGEMENT, version=2)
    )
    svc_families.families.append(
        DIBSuppSVCFamilies.Family(name=DIBServiceFamily.TUNNELING, version=1)
    )
    svc_families.families.append(
        DIBSuppSVCFamilies.Family(name=DIBServiceFamily.ROUTING, version=1)
    )
    svc_families.families.append(
        DIBSuppSVCFamilies.Family(
            name=DIBServiceFamily.REMOTE_CONFIGURATION_DIAGNOSIS, version=1
        )
    )

    frame_body.dibs.append(device_information)
    frame_body.dibs.append(svc_families)

    return KNXIPFrame.init_from_body(frame_body)



"""Abstraction to send ConnectonStateRequest and wait for ConnectionStateResponse."""

from __future__ import annotations

from typing import TYPE_CHECKING

from xknx.io.const import CONNECTIONSTATE_REQUEST_TIMEOUT
from xknx.knxip import HPAI, ConnectionStateRequest, ConnectionStateResponse, KNXIPFrame

from .request_response import RequestResponse

if TYPE_CHECKING:
    from xknx.io.transport import KNXIPTransport


class ConnectionState(RequestResponse):
    """Class to send ConnectonStateRequest and wait for ConnectionStateResponse."""

    def __init__(
        self,
        transport: KNXIPTransport,
        communication_channel_id: int,
        local_hpai: HPAI,
    ) -> None:
        """Initialize ConnectionState class."""
        super().__init__(
            transport,
            ConnectionStateResponse,
            timeout_in_seconds=CONNECTIONSTATE_REQUEST_TIMEOUT,
        )
        self.communication_channel_id = communication_channel_id
        self.local_hpai = local_hpai

    def create_knxipframe(self) -> KNXIPFrame:
        """Create KNX/IP Frame object to be sent to device."""
        connectionstate_request = ConnectionStateRequest(
            communication_channel_id=self.communication_channel_id,
            control_endpoint=self.local_hpai,
        )
        return KNXIPFrame.init_from_body(connectionstate_request)



"""Tests for the CEMIFrame object."""

import pytest

from xknx.cemi import (
    CEMIFlags,
    CEMIFrame,
    CEMILData,
    CEMIMessageCode,
    CEMIMPropReadRequest,
    CEMIMPropReadResponse,
    CEMIMPropWriteRequest,
    CEMIMPropWriteResponse,
)
from xknx.cemi.const import CEMIErrorCode
from xknx.exceptions import ConversionError, CouldNotParseCEMI, UnsupportedCEMIMessage
from xknx.profile.const import ResourceKNXNETIPPropertyId, ResourceObjectType
from xknx.telegram import GroupAddress, IndividualAddress, Telegram
from xknx.telegram.apci import GroupValueRead
from xknx.telegram.tpci import TConnect, TDataBroadcast, TDataGroup


def get_data(
    code: int,
    adil: int,
    flags: int,
    src: int,
    dst: int,
    npdu_len: int,
    tpci_apci: int,
    payload: list[int],
) -> bytes:
    """Encode to cemi data raw bytes."""
    return bytes(
        [
            code,
            adil,  # adil
            (flags >> 8) & 255,  # flags
            flags & 255,  # flags
            (src >> 8) & 255,  # src
            src & 255,  # src
            (dst >> 8) & 255,  # dst
            dst & 255,  # dst
            npdu_len,  # npdu_len
            (tpci_apci >> 8) & 255,  # tpci_apci
            tpci_apci & 255,  # tpci_apci
            *payload,  # payload
        ]
    )


def test_valid_command() -> None:
    """Test for valid frame parsing."""
    raw = get_data(0x29, 0, 0x0080, 1, 1, 1, 0, [])
    frame = CEMIFrame.from_knx(raw)
    assert frame.code == CEMIMessageCode.L_DATA_IND
    assert isinstance(frame.data, CEMILData)
    assert frame.data.flags == 0x0080
    assert frame.data.hops == 0
    assert frame.data.src_addr == IndividualAddress(1)
    assert frame.data.dst_addr == GroupAddress(1)
    assert frame.data.payload == GroupValueRead()
    assert frame.data.tpci == TDataGroup()
    assert frame.calculated_length() == 11
    assert frame.to_knx() == raw


def test_valid_tpci_control() -> None:
    """Test for valid tpci control."""
    raw = bytes((0x29, 0, 0, 0, 0, 0, 0, 0, 0, 0x80))
    frame = CEMIFrame.from_knx(raw)
    assert frame.code == CEMIMessageCode.L_DATA_IND
    assert isinstance(frame.data, CEMILData)
    assert frame.data.flags == 0
    assert frame.data.hops == 0
    assert frame.data.payload is None
    assert frame.data.src_addr == IndividualAddress(0)
    assert frame.data.dst_addr == IndividualAddress(0)
    assert frame.data.tpci == TConnect()
    assert frame.calculated_length() == 10
    assert frame.to_knx() == raw


@pytest.mark.parametrize(
    "raw,err_msg",
    [
        (
            get_data(0x29, 0, 0, 0, 0, 1, 0xFFC0, []),
            r".*Invalid length for control TPDU.*",
        ),
    ],
)
def test_invalid_tpci_apci(raw: bytes, err_msg: str) -> None:
    """Test for invalid APCIService."""
    with pytest.raises(CouldNotParseCEMI, match=err_msg):
        CEMIFrame.from_knx(raw)


@pytest.mark.parametrize(
    "raw,err_msg",
    [
        (
            get_data(0x29, 0, 0, 0, 0, 1, 0x08C0, []),
            r".*TPCI not supported.*",
        ),
        (
            get_data(0x29, 0, 0, 0, 0, 1, 0x03C0, []),
            r".*APDU not supported*",
        ),
    ],
)
def test_unsupported_tpci_apci(raw: bytes, err_msg: str) -> None:
    """Test for invalid APCIService."""
    with pytest.raises(UnsupportedCEMIMessage, match=err_msg):
        CEMIFrame.from_knx(raw)


def test_invalid_apdu_len() -> None:
    """Test for invalid apdu len."""
    with pytest.raises(CouldNotParseCEMI, match=r".*APDU LEN should be .*"):
        CEMIFrame.from_knx(get_data(0x29, 0, 0, 0, 0, 2, 0, []))


def test_invalid_payload() -> None:
    """Test for having wrong payload set."""
    frame = CEMIFrame(
        code=CEMIMessageCode.L_DATA_IND,
        data=CEMILData(
            flags=0,
            src_addr=IndividualAddress(0),
            dst_addr=IndividualAddress(0),
            tpci=TDataGroup(),
            payload=None,
        ),
    )

    with pytest.raises(TypeError):
        frame.calculated_length()

    with pytest.raises(ConversionError):
        frame.to_knx()


def test_missing_data() -> None:
    """Test for having no data set."""
    frame = CEMIFrame(
        code=CEMIMessageCode.L_DATA_IND,
        data=None,
    )

    with pytest.raises(UnsupportedCEMIMessage):
        frame.calculated_length()

    with pytest.raises(UnsupportedCEMIMessage):
        frame.to_knx()


def test_from_knx_with_not_handleable_cemi() -> None:
    """Test for having unhandlebale cemi set."""
    with pytest.raises(
        UnsupportedCEMIMessage, match=r".*CEMIMessageCode not implemented:.*"
    ):
        CEMIFrame.from_knx(get_data(0x30, 0, 0, 0, 0, 2, 0, []))


def test_from_knx_with_not_implemented_cemi() -> None:
    """Test for having not implemented CEMI set."""
    with pytest.raises(
        UnsupportedCEMIMessage, match=r".*Could not handle CEMIMessageCode:.*"
    ):
        CEMIFrame.from_knx(
            get_data(CEMIMessageCode.L_BUSMON_IND.value, 0, 0, 0, 0, 2, 0, [])
        )


def test_invalid_invalid_len() -> None:
    """Test for invalid cemi len."""
    with pytest.raises(CouldNotParseCEMI, match=r".*CEMI too small.*"):
        CEMIFrame.from_knx(get_data(0x29, 0, 0, 0, 0, 2, 0, [])[:5])


def test_from_knx_group_address() -> None:
    """Test conversion for a cemi with a group address as destination."""
    frame = CEMIFrame.from_knx(get_data(0x29, 0, 0x80, 0, 0, 1, 0, []))
    assert isinstance(frame.data, CEMILData)
    assert frame.data.dst_addr == GroupAddress(0)


def test_from_knx_individual_address() -> None:
    """Test conversion for a cemi with a individual address as destination."""
    frame = CEMIFrame.from_knx(get_data(0x29, 0, 0x00, 0, 0, 1, 0, []))
    assert isinstance(frame.data, CEMILData)
    assert frame.data.dst_addr == IndividualAddress(0)


def test_telegram_group_address() -> None:
    """Test telegram conversion flags with a group address."""
    _telegram = Telegram(destination_address=GroupAddress(1))
    frame = CEMIFrame(
        code=CEMIMessageCode.L_DATA_IND,
        data=CEMILData.init_from_telegram(_telegram),
    )
    assert isinstance(frame.data, CEMILData)
    assert frame.data.flags & 0x0080 == CEMIFlags.DESTINATION_GROUP_ADDRESS
    assert frame.data.flags & 0x0C00 == CEMIFlags.PRIORITY_LOW
    # test CEMIFrame.telegram property
    assert frame.data.telegram() == _telegram


def test_telegram_broadcast() -> None:
    """Test telegram conversion flags with a group address."""
    _telegram = Telegram(destination_address=GroupAddress(0))
    frame = CEMIFrame(
        code=CEMIMessageCode.L_DATA_IND,
        data=CEMILData.init_from_telegram(_telegram),
    )
    assert isinstance(frame.data, CEMILData)
    assert frame.data.flags & 0x0080 == CEMIFlags.DESTINATION_GROUP_ADDRESS
    assert frame.data.flags & 0x0C00 == CEMIFlags.PRIORITY_SYSTEM
    assert frame.data.tpci == TDataBroadcast()
    # test CEMIFrame.telegram property
    assert frame.data.telegram() == _telegram


def test_telegram_individual_address() -> None:
    """Test telegram conversion flags with a individual address."""
    _telegram = Telegram(destination_address=IndividualAddress(0), tpci=TConnect())
    frame = CEMIFrame(
        code=CEMIMessageCode.L_DATA_IND,
        data=CEMILData.init_from_telegram(_telegram),
    )
    assert isinstance(frame.data, CEMILData)
    assert frame.data.flags & 0x0080 == CEMIFlags.DESTINATION_INDIVIDUAL_ADDRESS
    assert frame.data.flags & 0x0C00 == CEMIFlags.PRIORITY_SYSTEM
    assert frame.data.flags & 0x0200 == CEMIFlags.NO_ACK_REQUESTED
    # test CEMIFrame.telegram property
    assert frame.data.telegram() == _telegram


def test_telegram_unsupported_address() -> None:
    """Test telegram conversion flags with an unsupported address."""
    with pytest.raises(TypeError):
        CEMIFrame(
            code=CEMIMessageCode.L_DATA_IND,
            data=CEMILData.init_from_telegram(Telegram(destination_address=object())),
        )


def get_prop(
    code: int,
    obj_id: int,
    obj_inst: int,
    prop_id: int,
    num: int,
    six: int,
    payload: list[int],
) -> bytes:
    """Encode to cemi prop raw bytes."""
    return bytes(
        [
            code,
            (obj_id >> 8) & 255,  # Interface Object Type
            obj_id & 255,  # Interface Object Type
            obj_inst & 255,  # Object instance
            prop_id & 255,  # Property ID
            (num << 4) | (six >> 8),  # Number of Elements (4bit) Start index (hsb 4bit)
            six & 255,  # Start index (lsb 8bit)
            *payload,  # payload
        ]
    )


def test_valid_read_req() -> None:
    """Test for valid frame parsing."""
    raw = get_prop(0xFC, 0x000B, 1, 52, 1, 1, [])
    frame = CEMIFrame.from_knx(raw)
    assert frame.code == CEMIMessageCode.M_PROP_READ_REQ
    assert isinstance(frame.data, CEMIMPropReadRequest)
    assert (
        frame.data.property_info.object_type
        == ResourceObjectType.OBJECT_KNXNETIP_PARAMETER
    )
    assert frame.data.property_info.object_instance == 1
    assert (
        frame.data.property_info.property_id
        == ResourceKNXNETIPPropertyId.PID_KNX_INDIVIDUAL_ADDRESS
    )
    assert frame.data.property_info.number_of_elements == 1
    assert frame.data.property_info.start_index == 1
    assert frame.calculated_length() == 7
    assert frame.to_knx() == raw
    with pytest.raises(AttributeError):
        frame.data.telegram()


def test_valid_read_con() -> None:
    """Test for valid frame parsing."""
    raw = get_prop(0xFB, 0x000B, 1, 52, 1, 1, [0x12, 0x03])
    frame = CEMIFrame.from_knx(raw)
    assert frame.code == CEMIMessageCode.M_PROP_READ_CON
    assert isinstance(frame.data, CEMIMPropReadResponse)
    assert (
        frame.data.property_info.object_type
        == ResourceObjectType.OBJECT_KNXNETIP_PARAMETER
    )
    assert frame.data.property_info.object_instance == 1
    assert (
        frame.data.property_info.property_id
        == ResourceKNXNETIPPropertyId.PID_KNX_INDIVIDUAL_ADDRESS
    )
    assert frame.data.property_info.number_of_elements == 1
    assert frame.data.property_info.start_index == 1
    assert frame.data.error_code is None
    assert IndividualAddress.from_knx(frame.data.data) == IndividualAddress("1.2.3")
    assert frame.calculated_length() == 9
    assert frame.to_knx() == raw


def test_valid_error_read_con() -> None:
    """Test for valid frame parsing."""
    raw = get_prop(0xFB, 0x000B, 1, 52, 0, 1, [0x07])
    frame = CEMIFrame.from_knx(raw)
    assert frame.code == CEMIMessageCode.M_PROP_READ_CON
    assert isinstance(frame.data, CEMIMPropReadResponse)
    assert (
        frame.data.property_info.object_type
        == ResourceObjectType.OBJECT_KNXNETIP_PARAMETER
    )
    assert frame.data.property_info.object_instance == 1
    assert (
        frame.data.property_info.property_id
        == ResourceKNXNETIPPropertyId.PID_KNX_INDIVIDUAL_ADDRESS
    )
    assert frame.data.property_info.number_of_elements == 0
    assert frame.data.property_info.start_index == 1
    assert frame.data.error_code == CEMIErrorCode.CEMI_ERROR_VOID_DP
    assert frame.calculated_length() == 8
    assert frame.to_knx() == raw


def test_valid_write_req() -> None:
    """Test for valid frame parsing."""
    raw = get_prop(0xF6, 0x000B, 1, 52, 1, 1, [0x12, 0x03])
    frame = CEMIFrame.from_knx(raw)
    assert frame.code == CEMIMessageCode.M_PROP_WRITE_REQ
    assert isinstance(frame.data, CEMIMPropWriteRequest)
    assert (
        frame.data.property_info.object_type
        == ResourceObjectType.OBJECT_KNXNETIP_PARAMETER
    )
    assert frame.data.property_info.object_instance == 1
    assert (
        frame.data.property_info.property_id
        == ResourceKNXNETIPPropertyId.PID_KNX_INDIVIDUAL_ADDRESS
    )
    assert frame.data.property_info.number_of_elements == 1
    assert frame.data.property_info.start_index == 1
    assert IndividualAddress.from_knx(frame.data.data) == IndividualAddress("1.2.3")
    assert frame.calculated_length() == 9
    assert frame.to_knx() == raw


def test_valid_empty_write_con() -> None:
    """Test for valid frame parsing."""
    raw = get_prop(0xF5, 0x000B, 1, 52, 1, 1, [])
    frame = CEMIFrame.from_knx(raw)
    assert frame.code == CEMIMessageCode.M_PROP_WRITE_CON
    assert isinstance(frame.data, CEMIMPropWriteResponse)
    assert (
        frame.data.property_info.object_type
        == ResourceObjectType.OBJECT_KNXNETIP_PARAMETER
    )
    assert frame.data.property_info.object_instance == 1
    assert (
        frame.data.property_info.property_id
        == ResourceKNXNETIPPropertyId.PID_KNX_INDIVIDUAL_ADDRESS
    )
    assert frame.data.property_info.number_of_elements == 1
    assert frame.data.property_info.start_index == 1
    assert frame.data.error_code is None
    assert frame.calculated_length() == 7
    assert frame.to_knx() == raw


def test_valid_error_write_con() -> None:
    """Test for valid frame parsing."""
    raw = get_prop(0xF5, 0x000B, 1, 52, 0, 1, [0x07])
    frame = CEMIFrame.from_knx(raw)
    assert frame.code == CEMIMessageCode.M_PROP_WRITE_CON
    assert isinstance(frame.data, CEMIMPropWriteResponse)
    assert (
        frame.data.property_info.object_type
        == ResourceObjectType.OBJECT_KNXNETIP_PARAMETER
    )
    assert frame.data.property_info.object_instance == 1
    assert (
        frame.data.property_info.property_id
        == ResourceKNXNETIPPropertyId.PID_KNX_INDIVIDUAL_ADDRESS
    )
    assert frame.data.property_info.number_of_elements == 0
    assert frame.data.property_info.start_index == 1
    assert frame.data.error_code == CEMIErrorCode.CEMI_ERROR_VOID_DP
    assert frame.calculated_length() == 8
    assert frame.to_knx() == raw


@pytest.mark.parametrize(
    "raw,err_msg",
    [
        (
            get_prop(0xFC, 0x000B, 1, 52, 1, 1, [])[:5],
            r".*Invalid CEMI length:*",
        ),
        (
            get_prop(0xFB, 0x000B, 1, 52, 1, 1, [])[:5],
            r".*CEMI Property Read Response too small.*",
        ),
        (
            get_prop(0xFB, 0x000B, 1, 52, 0, 1, [0x07, 0x00]),
            r".*Invalid CEMI error response length:.*",
        ),
        (
            get_prop(0xF6, 0x000B, 1, 52, 1, 1, [])[:5],
            r".*CEMI Property Write Request too small.*",
        ),
        (
            get_prop(0xF5, 0x000B, 1, 52, 1, 1, [])[:5],
            r".*CEMI Property Write Response too small.*",
        ),
        (
            get_prop(0xF5, 0x000B, 1, 52, 0, 1, [0x07, 0x00]),
            r".*Invalid CEMI error response length:.*",
        ),
        (
            get_prop(0xF5, 0x000B, 1, 52, 1, 1, [0x07]),
            r".*Invalid CEMI response length:.*",
        ),
    ],
)
def test_invalid_length(raw: bytes, err_msg: str) -> None:
    """Test for invalid frame parsing."""
    with pytest.raises(CouldNotParseCEMI, match=err_msg):
        CEMIFrame.from_knx(raw)


def test_invalid_resource_object() -> None:
    """Test for invalid frame parsing."""
    with pytest.raises(
        UnsupportedCEMIMessage, match=r".*CEMIMProp Object Type not supported:.*"
    ):
        CEMIFrame.from_knx(get_prop(0xFC, 0x1234, 1, 52, 1, 1, []))



"""
Module for Serialization and Deserialization of KNX Description Requests.

The DESCRIPTION_REQUEST frame shall be sent by the KNXnet/IP Client to the control endpoint
of the KNXnet/IP Server to obtain a self-description of the KNXnet/IP Server device.
"""

from __future__ import annotations

from .body import KNXIPBody
from .hpai import HPAI
from .knxip_enum import KNXIPServiceType


class DescriptionRequest(KNXIPBody):
    """Representation of a KNX Description Request."""

    SERVICE_TYPE = KNXIPServiceType.DESCRIPTION_REQUEST

    def __init__(self, control_endpoint: HPAI | None = None) -> None:
        """Initialize SearchRequest object."""
        self.control_endpoint = control_endpoint or HPAI()

    def calculated_length(self) -> int:
        """Get length of KNX/IP body."""
        return HPAI.LENGTH

    def from_knx(self, raw: bytes) -> int:
        """Parse/deserialize from KNX/IP raw data."""
        return self.control_endpoint.from_knx(raw)

    def to_knx(self) -> bytes:
        """Serialize to KNX/IP raw data."""
        return self.control_endpoint.to_knx()

    def __repr__(self) -> str:
        """Return object as readable string."""
        return f'<DescriptionRequest control_endpoint="{self.control_endpoint}" />'



"""Tests for KNX/IP routing connections."""

import asyncio
from unittest.mock import AsyncMock, Mock, call, patch

from xknx import XKNX
from xknx.cemi import CEMIFrame, CEMILData, CEMIMessageCode
from xknx.io import Routing
from xknx.io.routing import (
    BUSY_DECREMENT_TIME,
    BUSY_INCREMENT_COOLDOWN,
    BUSY_RANDOM_TIME_FACTOR,
    BUSY_SLOWDURATION_TIME_FACTOR,
    ROUTING_INDICATION_WAIT_TIME,
    _RoutingFlowControl,
)
from xknx.knxip import KNXIPFrame, RoutingBusy, RoutingIndication
from xknx.telegram import IndividualAddress, Telegram, TelegramDirection, tpci

from ..conftest import EventLoopClockAdvancer


class TestRouting:
    """Test class for xknx/io/Routing objects."""

    def setup_method(self) -> None:
        """Set up test class."""
        # pylint: disable=attribute-defined-outside-init
        self.xknx = XKNX()
        self.cemi_received_mock = AsyncMock()

    @patch("xknx.io.Routing._send_knxipframe")
    async def test_request_received_callback(self, send_knxipframe_mock: Mock) -> None:
        """Test Routing for responding to a transport layer connection."""
        routing = Routing(
            self.xknx,
            individual_address=None,
            cemi_received_callback=self.xknx.knxip_interface.cemi_received,
            local_ip="192.168.1.1",
        )
        self.xknx.knxip_interface._interface = routing
        # set current address so management telegram is processed
        self.xknx.current_address = IndividualAddress("1.0.255")
        # L_Data.ind T_Connect from 1.0.250 to 1.0.255 (xknx tunnel endpoint)
        # communication_channel_id: 0x02   sequence_counter: 0x81
        raw_ind = bytes.fromhex("0610 0530 0010 2900b06010fa10ff0080")
        _cemi = CEMIFrame.from_knx(raw_ind[6:])
        test_telegram = _cemi.data.telegram()
        test_telegram.direction = TelegramDirection.INCOMING

        response_telegram = Telegram(
            destination_address=IndividualAddress(test_telegram.source_address),
            tpci=tpci.TDisconnect(),
        )
        response_cemi = CEMIFrame(
            code=CEMIMessageCode.L_DATA_IND,
            data=CEMILData.init_from_telegram(
                telegram=response_telegram,
                src_addr=IndividualAddress("1.0.255"),
            ),
        )
        response_frame = KNXIPFrame.init_from_body(
            RoutingIndication(raw_cemi=response_cemi.to_knx())
        )

        routing.transport.data_received_callback(raw_ind, ("192.168.1.2", 3671))
        await asyncio.sleep(0)
        assert send_knxipframe_mock.call_args_list == [
            call(response_frame),
        ]
        await asyncio.sleep(0)  # await local L_Data.con

    @patch("logging.Logger.warning")
    async def test_routing_lost_message(self, logging_mock: Mock) -> None:
        """Test class for received RoutingLostMessage frames."""
        routing = Routing(
            self.xknx,
            individual_address=None,
            cemi_received_callback=AsyncMock(),
            local_ip="192.168.1.1",
        )
        raw = bytes((0x06, 0x10, 0x05, 0x31, 0x00, 0x0A, 0x04, 0x00, 0x00, 0x05))
        routing.transport.data_received_callback(raw, ("192.168.1.2", 3671))
        logging_mock.assert_called_once_with(
            "RoutingLostMessage received from %s - %s lost messages.",
            "192.168.1.2",
            5,
        )


class TestFlowControl:
    """Test class for KNXnet/IP routing flow control."""

    async def test_basic_throttling(self, time_travel: EventLoopClockAdvancer) -> None:
        """Test throttling outgoing frames."""
        flow_control = _RoutingFlowControl()
        mock = Mock()

        async def test_send() -> None:
            async with flow_control.throttle():
                mock()

        # first send is called immediately
        task = asyncio.create_task(test_send())
        await asyncio.sleep(0)
        assert mock.call_count == 1
        assert task.done()
        mock.reset_mock()

        # second send is throttled
        task = asyncio.create_task(test_send())
        await asyncio.sleep(0)
        assert mock.call_count == 0
        await time_travel(ROUTING_INDICATION_WAIT_TIME / 4)
        assert not task.done()
        await time_travel(ROUTING_INDICATION_WAIT_TIME / 4 * 3)
        assert task.done()
        assert mock.call_count == 1
        mock.reset_mock()

        # later send is called immediately
        await time_travel(ROUTING_INDICATION_WAIT_TIME)
        task = asyncio.create_task(test_send())
        await asyncio.sleep(0)
        assert mock.call_count == 1
        assert task.done()
        mock.reset_mock()

    @patch("random.random")
    async def test_routing_busy(
        self, random_mock: Mock, time_travel: EventLoopClockAdvancer
    ) -> None:
        """Test throttling on received RoutingBusy frame."""
        flow_control = _RoutingFlowControl()
        mock = Mock()
        test_wait_time_ms = 100
        random_mock.return_value = 0.5

        async def test_send() -> None:
            async with flow_control.throttle():
                mock()

        test_busy = RoutingBusy(wait_time=test_wait_time_ms)

        flow_control.handle_routing_busy(test_busy)
        task = asyncio.create_task(test_send())
        await asyncio.sleep(0)
        assert mock.call_count == 0
        await time_travel(test_wait_time_ms / 1000)
        assert mock.call_count == 1
        assert task.done()
        # no slowduration for just 1 RoutingBusy
        assert flow_control._timer_task.done()
        mock.reset_mock()

        # multiple RoutingBusy frames
        flow_control.handle_routing_busy(test_busy)
        # after cooldown - with different wait times updating wait time for 2x time
        # not counting one frame due to cooldown time
        await time_travel(BUSY_INCREMENT_COOLDOWN)
        flow_control.handle_routing_busy(RoutingBusy(wait_time=test_wait_time_ms // 2))
        assert flow_control._received_busy_frames == 1
        assert flow_control._wait_time_ms == test_wait_time_ms
        # doesn't count in same 10 ms window but updates wait time
        flow_control.handle_routing_busy(RoutingBusy(wait_time=test_wait_time_ms * 2))
        assert flow_control._received_busy_frames == 1
        assert flow_control._wait_time_ms == test_wait_time_ms * 2
        # add second busy frame after cooldown has passed
        await time_travel(BUSY_INCREMENT_COOLDOWN)
        flow_control.handle_routing_busy(RoutingBusy(wait_time=test_wait_time_ms // 2))
        assert flow_control._wait_time_ms == test_wait_time_ms * 2
        assert flow_control._received_busy_frames == 2

        task = asyncio.create_task(test_send())
        assert mock.call_count == 0
        await time_travel(test_wait_time_ms * 2 / 1000 - BUSY_INCREMENT_COOLDOWN)
        assert mock.call_count == 0  # not yet resolved, wait for additional random time
        await time_travel(
            flow_control._received_busy_frames * BUSY_RANDOM_TIME_FACTOR * random_mock()
        )
        assert mock.call_count == 1
        assert task.done()
        # slowduration
        assert not flow_control._timer_task.done()
        await time_travel(2 * BUSY_SLOWDURATION_TIME_FACTOR)  # _received_busy_frames 2
        await time_travel(BUSY_DECREMENT_TIME)  # and decrement time
        assert not flow_control._timer_task.done()
        await time_travel(BUSY_DECREMENT_TIME)  # and second decrement time
        assert flow_control._timer_task.done()



"""Implementation of the KNX datetime data point."""

from __future__ import annotations

from collections.abc import Mapping
from dataclasses import asdict, dataclass
import datetime
from typing import Any

from xknx.exceptions import ConversionError

from .dpt import DPTComplex, DPTComplexData, DPTEnumData
from .payload import DPTArray, DPTBinary


class KNXDayOfWeek(DPTEnumData):
    """Enum for the different KNX days."""

    ANY_DAY = 0
    MONDAY = 1
    TUESDAY = 2
    WEDNESDAY = 3
    THURSDAY = 4
    FRIDAY = 5
    SATURDAY = 6
    SUNDAY = 7


@dataclass(slots=True)
class KNXDateTime(DPTComplexData):
    """
    Class for KNX DateTime.

    `year`, `day_of_week` and `working_day` may be None if invalid.
    `month` and `day` have to be both either int or None (invalid).
    `hour`, `minutes` and `seconds` all have to be either int or None (invalid).
    """

    year: int | None = None
    month: int | None = None
    day: int | None = None
    hour: int | None = None
    minutes: int | None = None
    seconds: int | None = None

    day_of_week: KNXDayOfWeek | None = None

    fault: bool = False
    working_day: bool | None = None
    dst: bool = False
    external_sync: bool = False
    source_reliable: bool = False

    @classmethod
    def from_dict(cls, data: Mapping[str, Any]) -> KNXDateTime:
        """Init from a dictionary."""
        _data = {**data}
        if "day_of_week" in data:
            _data["day_of_week"] = KNXDayOfWeek.parse(data["day_of_week"])
        try:
            return cls(**_data)
        except TypeError as err:
            raise ValueError(f"Invalid value for KNXDateTime: {err}") from err

    def as_dict(self) -> dict[str, int | str]:
        """Create a JSON serializable dictionary."""
        _data = asdict(self)
        if self.day_of_week is not None:
            _data["day_of_week"] = self.day_of_week.name.lower()
        return _data

    def as_datetime(self) -> datetime.datetime:
        """Return datetime object."""
        try:
            return datetime.datetime(
                self.year,  # type: ignore[arg-type]
                self.month,  # type: ignore[arg-type]
                self.day,  # type: ignore[arg-type]
                self.hour,  # type: ignore[arg-type]
                self.minutes,  # type: ignore[arg-type]
                self.seconds,  # type: ignore[arg-type]
            )
        except (TypeError, ValueError) as err:
            raise ValueError(
                f"KNXDateTime not convertible to datetime object: {err}"
            ) from err

    @classmethod
    def from_datetime(cls, dt: datetime.datetime) -> KNXDateTime:
        """Return KNXDateTime object from a datetime.datetime object."""
        return cls(
            year=dt.year,
            month=dt.month,
            day=dt.day,
            hour=dt.hour,
            minutes=dt.minute,
            seconds=dt.second,
        )


class DPTDateTime(DPTComplex[KNXDateTime]):
    """Abstraction for KNX 8 octet datetime (DPT 19.001)."""

    data_type = KNXDateTime
    payload_type = DPTArray
    payload_length = 8
    dpt_main_number = 19
    dpt_sub_number = 1
    value_type = "datetime"

    @classmethod
    def from_knx(cls, payload: DPTArray | DPTBinary) -> KNXDateTime:
        """Parse/deserialize from KNX/IP raw data."""
        raw = cls.validate_payload(payload)

        year = raw[0] + 1900
        month = raw[1] & 0b00001111
        day = raw[2] & 0b00011111
        weekday = (raw[3] & 0b11100000) >> 5
        hour = raw[3] & 0b00011111
        minutes = raw[4] & 0b00111111
        seconds = raw[5] & 0b00111111

        fault = bool(raw[6] & 0b10000000)
        _working_day = bool(raw[6] & 0b01000000)
        _working_day_invalid = bool(raw[6] & 0b00100000)
        _year_invalid = bool(raw[6] & 0b00010000)
        _date_invalid = bool(raw[6] & 0b00001000)  # month, day
        _weekday_invalid = bool(raw[6] & 0b00000100)
        _time_invalid = bool(raw[6] & 0b00000010)  # hour, minutes, seconds
        dst = bool(raw[6] & 0b00000001)

        external_sync = bool(raw[7] & 0b10000000)
        source_reliable = bool(raw[7] & 0b01000000)

        knx_date_time = KNXDateTime(
            year=year if not _year_invalid else None,
            month=month if not _date_invalid else None,
            day=day if not _date_invalid else None,
            day_of_week=KNXDayOfWeek(weekday) if not _weekday_invalid else None,
            hour=hour if not _time_invalid else None,
            minutes=minutes if not _time_invalid else None,
            seconds=seconds if not _time_invalid else None,
            fault=fault,
            working_day=_working_day if not _working_day_invalid else None,
            dst=dst,
            external_sync=external_sync,
            source_reliable=source_reliable,
        )
        try:
            cls._test_range(knx_date_time)
        except ValueError as err:
            raise ConversionError(f"Could not parse {cls.dpt_name()}: {err}") from err
        return knx_date_time

    @classmethod
    def _to_knx(cls, value: KNXDateTime) -> DPTArray:
        """Serialize to KNX/IP raw data."""
        try:
            cls._test_range(value)
        except ValueError as err:
            raise ConversionError(
                f"Could not serialize {cls.dpt_name()}: {err}"
            ) from err

        knx_year = (value.year - 1900) & 0xFF if value.year is not None else 0

        month_day_invalid = None in (value.month, value.day)
        month = value.month if value.month is not None else 0
        day = value.day if value.day is not None else 0

        day_of_week = value.day_of_week.value if value.day_of_week is not None else 0

        time_invalid = None in (value.hour, value.minutes, value.seconds)
        hour = value.hour if value.hour is not None else 0
        minutes = value.minutes if value.minutes is not None else 0
        seconds = value.seconds if value.seconds is not None else 0

        return DPTArray(
            (
                knx_year,
                month,
                day,
                day_of_week << 5 | hour,
                minutes,
                seconds,
                (
                    value.fault << 7
                    | bool(value.working_day) << 6
                    | (value.working_day is None) << 5
                    | (value.year is None) << 4
                    | month_day_invalid << 3
                    | (value.day_of_week is None) << 2
                    | time_invalid << 1
                    | value.dst
                ),
                (value.external_sync << 7 | value.source_reliable << 6),
            )
        )

    @staticmethod
    def _test_range(value: KNXDateTime) -> None:
        """Test if date is in valid range."""
        if value.year is not None and not 1900 <= value.year <= 2155:
            raise ValueError(f"Year out of range 1900..2155: {value.year}")
        if value.month is not None and not 1 <= value.month <= 12:
            raise ValueError(f"Month out of range 1..12: {value.month}")
        if value.day is not None and not 1 <= value.day <= 31:
            raise ValueError(f"Day out of range 1..31: {value.day}")
        if value.hour is not None and not 0 <= value.hour <= 24:
            raise ValueError(f"Hour out of range 0..24: {value.hour}")
        if value.minutes is not None and not 0 <= value.minutes <= 59:
            raise ValueError(f"Minutes out of range 0..59: {value.minutes}")
        if value.seconds is not None and not 0 <= value.seconds <= 59:
            raise ValueError(f"Seconds out of range 0..59: {value.seconds}")
        if value.hour == 24 and (value.minutes != 0 or value.seconds != 0):
            raise ValueError(
                "Invalid time. When hour is 24, minutes and seconds have to be set to zero."
            )



"""Unit test for KNX color objects."""

from typing import Any

import pytest

from xknx.dpt import (
    DPTArray,
    DPTColorRGB,
    DPTColorRGBW,
    DPTColorXYY,
    RGBColor,
    RGBWColor,
    XYYColor,
)
from xknx.exceptions import ConversionError, CouldNotParseTelegram


class TestRGBColor:
    """Test RGBColor class."""

    @pytest.mark.parametrize(
        ("data", "value"),
        [
            ({"red": 0, "green": 128, "blue": 255}, RGBColor(0, 128, 255)),
            ({"red": 255, "green": 255, "blue": 255}, RGBColor(255, 255, 255)),
            ({"red": 0, "green": 0, "blue": 0}, RGBColor(0, 0, 0)),
        ],
    )
    def test_dict(self, data: dict[str, int], value: RGBColor) -> None:
        """Test from_dict and as_dict methods."""
        test_value = RGBColor.from_dict(data)
        assert test_value == value
        assert value.as_dict() == data

    @pytest.mark.parametrize(
        "data",
        [
            # incomplete data
            {},
            {"red": 128},
            {"green": 128},
            {"blue": 128},
            {"red": 128, "green": 128},
            {"red": 128, "blue": 128},
            {"green": 128, "blue": 128},
            # invalid data
            {"red": "a", "green": 128, "blue": 128},
            {"red": 128, "green": "a", "blue": 128},
            {"red": 128, "green": 128, "blue": "a"},
        ],
    )
    def test_dict_invalid(self, data: dict[str, Any]) -> None:
        """Test from_dict and as_dict methods."""
        with pytest.raises(ValueError):
            RGBColor.from_dict(data)


class TestDPTColorRGB:
    """Test class for KNX RGB-color objects."""

    @pytest.mark.parametrize(
        ("value", "raw"),
        [
            (RGBColor(0, 0, 0), (0x00, 0x00, 0x00)),  # min values
            (RGBColor(255, 255, 255), (0xFF, 0xFF, 0xFF)),  # max values
            (RGBColor(128, 128, 128), (0x80, 0x80, 0x80)),  # mid values
        ],
    )
    def test_rgbcolor_value(self, value: RGBColor, raw: tuple[int]) -> None:
        """Test DPTColorRGB parsing and streaming."""
        knx_value = DPTColorRGB.to_knx(value)
        assert knx_value == DPTArray(raw)
        assert DPTColorRGB.from_knx(knx_value) == value

    @pytest.mark.parametrize(
        ("value", "raw"),
        [
            ({"red": 128, "green": 128, "blue": 128}, (0x80, 0x80, 0x80)),
            ({"red": 255, "green": 255, "blue": 255}, (0xFF, 0xFF, 0xFF)),
        ],
    )
    def test_rgbcolor_to_knx_from_dict(
        self, value: dict[str, int], raw: tuple[int]
    ) -> None:
        """Test DPTColorRGB parsing from a dict."""
        knx_value = DPTColorRGB.to_knx(value)
        assert knx_value == DPTArray(raw)

    @pytest.mark.parametrize(
        ("red", "green", "blue"),
        [
            (-1, 0, 0),
            (0, -1, 0),
            (0, 0, -1),
            (256, 0, 0),
            (0, 256, 0),
            (0, 0, 256),
        ],
    )
    def test_rgbcolor_to_knx_limits(self, red: int, green: int, blue: int) -> None:
        """Test initialization of DPTColorRGB with wrong value."""
        value = RGBColor(red=red, green=green, blue=blue)
        with pytest.raises(ConversionError):
            DPTColorRGB.to_knx(value)

    @pytest.mark.parametrize(
        "value",
        [
            None,
            (0xFF, 0x4E, 0x12),
            1,
            ((0x00, 0xFF, 0x4E), 0x12),
            ((0xFF, 0x4E), (0x12, 0x00)),
            ((0, 0), "a"),
            ((0, 0), 0.4),
            RGBColor(red=0, green=0, blue="a"),
            RGBColor(red=4, green="a", blue=4),
            RGBColor(red="a", green=0, blue=1),
        ],
    )
    def test_rgbcolor_wrong_value_to_knx(self, value: Any) -> None:
        """Test DPTColorRGB parsing with wrong value."""
        with pytest.raises(ConversionError):
            DPTColorRGB.to_knx(value)

    def test_rgbcolor_wrong_value_from_knx(self) -> None:
        """Test DPTColorRGB parsing with wrong value."""
        with pytest.raises(CouldNotParseTelegram):
            DPTColorRGB.from_knx(DPTArray((0xFF, 0x4E, 0x12, 0x23)))
        with pytest.raises(CouldNotParseTelegram):
            DPTColorRGB.from_knx(DPTArray((0xFF, 0x4E)))


class TestRGBWColor:
    """Test RGBWColor class."""

    @pytest.mark.parametrize(
        ("data", "value"),
        [
            (
                {"red": 128, "green": 128, "blue": 128, "white": 128},
                RGBWColor(128, 128, 128, 128),
            ),
            (
                {"red": 255, "green": 255, "blue": 255, "white": 255},
                RGBWColor(255, 255, 255, 255),
            ),
            ({"red": 128, "green": 128, "blue": 128}, RGBWColor(128, 128, 128)),
            ({"white": 50}, RGBWColor(None, None, None, 50)),
            ({}, RGBWColor()),
            (
                {"red": None, "green": None, "blue": None, "white": 255},
                RGBWColor(None, None, None, 255),
            ),
            (
                {"red": 128, "green": 128, "blue": 128, "white": None},
                RGBWColor(128, 128, 128, None),
            ),
            ({"red": None, "white": 255}, RGBWColor(None, None, None, 255)),
            ({"green": None, "white": 255}, RGBWColor(None, None, None, 255)),
            ({"red": 128}, RGBWColor(128, None, None, None)),
            ({"green": 128}, RGBWColor(None, 128, None, None)),
            ({"blue": 128}, RGBWColor(None, None, 128, None)),
            ({"red": 128, "white": 128}, RGBWColor(128, None, None, 128)),
            ({"green": 128, "white": 128}, RGBWColor(None, 128, None, 128)),
            ({"blue": 128, "white": 128}, RGBWColor(None, None, 128, 128)),
        ],
    )
    def test_dict(self, data: dict[str, int], value: RGBWColor) -> None:
        """Test from_dict and as_dict methods."""
        test_value = RGBWColor.from_dict(data)
        assert test_value == value
        # fields default to `None`
        default_dict = {"red": None, "green": None, "blue": None, "white": None}
        assert value.as_dict() == default_dict | data

    @pytest.mark.parametrize(
        "data",
        [
            # invalid data
            {"red": "a", "green": 128, "blue": 128, "white": 128},
            {"red": 128, "green": "a", "blue": 128, "white": 128},
            {"red": 128, "green": 128, "blue": "a", "white": 128},
            {"red": 128, "green": 128, "blue": 128, "white": "a"},
        ],
    )
    def test_dict_invalid(self, data: dict[str, Any]) -> None:
        """Test from_dict and as_dict methods."""
        with pytest.raises(ValueError):
            RGBWColor.from_dict(data)

    def test_merge(self) -> None:
        """Test merging two RGBWColor objects."""
        color1 = RGBWColor(1, 1, 1, None)
        color2 = RGBWColor(None, None, None, 2)
        color3 = RGBWColor(3, None, None, 3)
        assert color1 | color2 == RGBWColor(1, 1, 1, 2)
        assert color2 | color1 == RGBWColor(1, 1, 1, 2)
        assert color1 | color3 == RGBWColor(3, 1, 1, 3)
        assert color3 | color1 == RGBWColor(1, 1, 1, 3)
        assert color2 | color3 == RGBWColor(3, None, None, 3)
        assert color3 | color2 == RGBWColor(3, None, None, 2)


class TestDPTColorRGBW:
    """Test class for KNX RGBW-color objects."""

    @pytest.mark.parametrize(
        ("value", "raw"),
        [
            (RGBWColor(0, 0, 0, 0), (0x00, 0x00, 0x00, 0x00, 0, 0xF)),  # min values
            (RGBWColor(255, 255, 255, 255), (0xFF, 0xFF, 0xFF, 0xFF, 0, 0xF)),  # max
            (RGBWColor(None, 0, 0, 0), (0x00, 0x00, 0x00, 0x00, 0, 0x7)),  # red None
            (RGBWColor(0, 0, 0, None), (0x00, 0x00, 0x00, 0x00, 0, 0xE)),  # white None
            (RGBWColor(None, None, None, None), (0x00, 0x00, 0x00, 0x00, 0, 0)),
            (RGBWColor(128, 128, 128, 128), (0x80, 0x80, 0x80, 0x80, 0, 0b1111)),
            (RGBWColor(204, 204, 204, 204), (0xCC, 0xCC, 0xCC, 0xCC, 0, 0b1111)),
        ],
    )
    def test_rgbwcolor_value(self, value: RGBWColor, raw: tuple[int]) -> None:
        """Test DPTColorRGBW parsing and streaming."""
        knx_value = DPTColorRGBW.to_knx(value)
        assert knx_value == DPTArray(raw)
        assert DPTColorRGBW.from_knx(knx_value) == value

    @pytest.mark.parametrize(
        ("value", "raw"),
        [
            (
                {"red": 128, "green": 128, "blue": 128, "white": 128},
                (0x80, 0x80, 0x80, 0x80, 0, 0xF),
            ),
            (
                {"red": 204, "green": 204, "blue": 204, "white": 204},
                (0xCC, 0xCC, 0xCC, 0xCC, 0, 0xF),
            ),
            (
                {"red": 204, "green": 204, "blue": 204},
                (0xCC, 0xCC, 0xCC, 0x00, 0, 0xE),
            ),
        ],
    )
    def test_rgbwcolor_to_knx_from_dict(
        self, value: dict[str, int], raw: tuple[int]
    ) -> None:
        """Test DPTColorRGBW parsing from a dict."""
        knx_value = DPTColorRGBW.to_knx(value)
        assert knx_value == DPTArray(raw)

    @pytest.mark.parametrize(
        ("red", "green", "blue", "white"),
        [
            (-1, 0, 0, 0),
            (0, -1, 0, 0),
            (0, 0, -1, 0),
            (0, 0, 0, -1),
            (256, 0, 0, 0),
            (0, 256, 0, 0),
            (0, 0, 256, 0),
            (0, 0, 0, 256),
        ],
    )
    def test_rgbwcolor_to_knx_limits(
        self, red: int, green: int, blue: int, white: int
    ) -> None:
        """Test initialization of DPTColorRGBW with wrong value."""
        value = RGBWColor(red=red, green=green, blue=blue, white=white)
        with pytest.raises(ConversionError):
            DPTColorRGBW.to_knx(value)

    @pytest.mark.parametrize(
        "value",
        [
            None,
            (0xFF, 0x4E, 0x12),
            1,
            ((0x00, 0xFF, 0x4E), 0x12),
            ((0xFF, 0x4E), (0x12, 0x00)),
            ((0, 0), "a"),
            ((0, 0), 0.4),
            RGBWColor(red=0, green=0, blue=0, white="a"),
            RGBWColor(red=4, green=4, blue="a", white=4),
            RGBWColor(red="a", green=0, blue=0, white=1),
        ],
    )
    def test_rgbwcolor_wrong_value_to_knx(self, value: Any) -> None:
        """Test DPTColorRGBW parsing with wrong value."""
        with pytest.raises(ConversionError):
            DPTColorRGBW.to_knx(value)

    def test_rgbwcolor_wrong_value_from_knx(self) -> None:
        """Test DPTColorRGBW parsing with wrong value."""
        with pytest.raises(CouldNotParseTelegram):
            DPTColorRGBW.from_knx(DPTArray((0xFF, 0x4E, 0x12)))


class TestXYYColor:
    """Test XYYColor class."""

    @pytest.mark.parametrize(
        ("data", "value"),
        [
            (
                {"x_axis": 0.1, "y_axis": 0.2, "brightness": 128},
                XYYColor((0.1, 0.2), 128),
            ),
            (
                {"x_axis": 0.5, "y_axis": 0.5, "brightness": 255},
                XYYColor((0.5, 0.5), 255),
            ),
            (
                {"x_axis": 0.9, "y_axis": 0.8},
                XYYColor((0.9, 0.8)),
            ),
            (
                {"brightness": 50},
                XYYColor(None, 50),
            ),
            (
                {},
                XYYColor(),
            ),
            (
                {"x_axis": None, "y_axis": None, "brightness": 255},
                XYYColor(None, 255),
            ),
            (
                {"x_axis": 0.5, "y_axis": 0.5, "brightness": None},
                XYYColor((0.5, 0.5), None),
            ),
            (
                {"x_axis": None, "brightness": 255},
                XYYColor(None, 255),
            ),
            (
                {"y_axis": None, "brightness": 255},
                XYYColor(None, 255),
            ),
        ],
    )
    def test_dict(self, data: dict[str, Any], value: XYYColor) -> None:
        """Test from_dict and as_dict methods."""
        test_value = XYYColor.from_dict(data)
        assert test_value == value
        # fields default to `None`
        default_dict = {"x_axis": None, "y_axis": None, "brightness": None}
        assert value.as_dict() == default_dict | data

    @pytest.mark.parametrize(
        "data",
        [
            # incomplete data
            {"x_axis": 0.1},
            {"y_axis": 0.1},
            {"x_axis": 0.1, "brightness": 128},
            {"y_axis": 0.1, "brightness": 128},
            # invalid data
            {"x_axis": "a", "y_axis": 0.1, "brightness": 128},
            {"x_axis": 0.1, "y_axis": "a", "brightness": 128},
            {"x_axis": 0.1, "y_axis": 0.1, "brightness": "a"},
        ],
    )
    def test_dict_invalid(self, data: dict[str, Any]) -> None:
        """Test from_dict and as_dict methods."""
        with pytest.raises(ValueError):
            XYYColor.from_dict(data)

    def test_merge(self) -> None:
        """Test merging two XYYColor objects."""
        color1 = XYYColor((1, 1), None)
        color2 = XYYColor(None, 2)
        color3 = XYYColor((3, 3), 3)
        assert color1 | color2 == XYYColor((1, 1), 2)
        assert color2 | color1 == XYYColor((1, 1), 2)
        assert color1 | color3 == XYYColor((3, 3), 3)
        assert color3 | color1 == XYYColor((1, 1), 3)
        assert color2 | color3 == XYYColor((3, 3), 3)
        assert color3 | color2 == XYYColor((3, 3), 2)


class TestDPTColorXYY:
    """Test class for KNX xyY-color objects."""

    @pytest.mark.parametrize(
        ("value", "raw"),
        [
            (XYYColor((0, 0), 0), (0x00, 0x00, 0x00, 0x00, 0x00, 0x03)),  # min values
            (XYYColor((1, 1), 255), (0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0x03)),  # max values
            (XYYColor(None, 0), (0x00, 0x00, 0x00, 0x00, 0x00, 0x01)),  # color None
            (XYYColor((0, 0), None), (0x00, 0x00, 0x00, 0x00, 0x00, 0x02)),
            (XYYColor(None, None), (0x00, 0x00, 0x00, 0x00, 0x00, 0x00)),
            (XYYColor((0.2, 0.2), 128), (0x33, 0x33, 0x33, 0x33, 0x80, 0x03)),
            (XYYColor((0.8, 0.8), 204), (0xCC, 0xCC, 0xCC, 0xCC, 0xCC, 0x03)),
        ],
    )
    def test_xyycolor_value(self, value: XYYColor, raw: tuple[int]) -> None:
        """Test DPTColorXYY parsing and streaming."""
        knx_value = DPTColorXYY.to_knx(value)
        assert knx_value == DPTArray(raw)
        assert DPTColorXYY.from_knx(knx_value) == value

    @pytest.mark.parametrize(
        ("value", "raw"),
        [
            (
                {"x_axis": 0.2, "y_axis": 0.2, "brightness": 128},
                (0x33, 0x33, 0x33, 0x33, 0x80, 0x03),
            ),
            (
                {"x_axis": 0.8, "y_axis": 0.8},
                (0xCC, 0xCC, 0xCC, 0xCC, 0x00, 0x02),
            ),
        ],
    )
    def test_xyycolor_to_knx_from_dict(
        self, value: dict[str, Any], raw: tuple[int]
    ) -> None:
        """Test DPTColorXYY parsing from a dict."""
        knx_value = DPTColorXYY.to_knx(value)
        assert knx_value == DPTArray(raw)

    @pytest.mark.parametrize(
        ("color", "brightness"),
        [
            ((-0.1, 0), 0),
            ((0, -0.1), 0),
            ((0, 0), -1),
            ((1.1, 0), 0),
            ((0, 1.1), 0),
            ((0, 0), 256),
        ],
    )
    def test_xyycolor_to_knx_limits(
        self, color: tuple[float, float], brightness: int
    ) -> None:
        """Test initialization of DPTColorXYY with wrong value."""
        value = XYYColor(color=color, brightness=brightness)
        with pytest.raises(ConversionError):
            DPTColorXYY.to_knx(value)

    @pytest.mark.parametrize(
        "value",
        [
            None,
            (0xFF, 0x4E, 0x12),
            1,
            ((0x00, 0xFF, 0x4E), 0x12),
            ((0xFF, 0x4E), (0x12, 0x00)),
            ((0, 0), "a"),
            ((0, 0), 0.4),
            XYYColor(color=(0, 0), brightness="a"),
            XYYColor(color=4, brightness=4),
            XYYColor(color=("a", 0), brightness=1),
        ],
    )
    def test_xyycolor_wrong_value_to_knx(self, value: Any) -> None:
        """Test DPTColorXYY parsing with wrong value."""
        with pytest.raises(ConversionError):
            DPTColorXYY.to_knx(value)

    def test_xyycolor_wrong_value_from_knx(self) -> None:
        """Test DPTColorXYY parsing with wrong value."""
        with pytest.raises(CouldNotParseTelegram):
            DPTColorXYY.from_knx(DPTArray((0xFF, 0x4E, 0x12)))



"""
CEMI Frame handler.

This class represents a CEMI Client vaguely according to KNX specification 3/6/3 §4.1.2.
It is responsible for sending and receiving CEMI frames to/from a CEMI Server - this
can be a remote server when using IP tunnelling or a local server when using IP routing.
"""

from __future__ import annotations

import asyncio
import logging
from typing import TYPE_CHECKING

from xknx.exceptions import (
    CommunicationError,
    ConfirmationError,
    ConversionError,
    CouldNotParseCEMI,
    DataSecureError,
    UnsupportedCEMIMessage,
)
from xknx.secure.data_secure import DataSecure, is_data_secure
from xknx.secure.keyring import Keyring
from xknx.telegram import IndividualAddress, Telegram, TelegramDirection, tpci
from xknx.util import asyncio_timeout

from .cemi_frame import CEMIFrame, CEMILData
from .const import CEMIMessageCode

if TYPE_CHECKING:
    from xknx.xknx import XKNX

logger = logging.getLogger("xknx.cemi")
data_secure_logger = logging.getLogger("xknx.data_secure")

# See 3/6/3 EMI_IMI §4.1.5 Data Link Layer messages
REQUEST_TO_CONFIRMATION_TIMEOUT = 3


class CEMIHandler:
    """Class for handling CEMI frames from/to the TelegramQueue."""

    __slots__ = ("_l_data_confirmation_event", "data_secure", "xknx")

    def __init__(self, xknx: XKNX) -> None:
        """Initialize CEMIHandler class."""
        self.xknx = xknx
        self.data_secure: DataSecure | None = None
        self._l_data_confirmation_event = asyncio.Event()

    def data_secure_init(self, keyring: Keyring | None) -> None:
        """Initialize DataSecure."""
        if keyring is None:
            self.data_secure = None
        else:
            self.data_secure = DataSecure.init_from_keyring(keyring)

    async def send_telegram(self, telegram: Telegram) -> None:
        """Create a CEMIFrame from a Telegram and send it to the CEMI Server."""
        cemi_data = CEMILData.init_from_telegram(
            telegram=telegram,
            src_addr=(
                self.xknx.current_address if telegram.source_address.raw == 0 else None
            ),
        )
        cemi = CEMIFrame(
            code=CEMIMessageCode.L_DATA_REQ,
            data=cemi_data,
        )

        logger.debug("Outgoing CEMI: %s", cemi)
        if self.data_secure is not None:
            cemi.data = self.data_secure.outgoing_cemi(cemi_data=cemi_data)
            telegram.data_secure = is_data_secure(cemi.data)
        else:
            telegram.data_secure = False

        self._l_data_confirmation_event.clear()
        try:
            await self.xknx.knxip_interface.send_cemi(cemi)
        except (ConversionError, CommunicationError) as ex:
            logger.warning("Could not send CEMI frame: %s for %s", ex, cemi)
            self.xknx.connection_manager.cemi_count_outgoing_error += 1
            raise ex

        try:
            async with asyncio_timeout(REQUEST_TO_CONFIRMATION_TIMEOUT):
                await self._l_data_confirmation_event.wait()
        except asyncio.TimeoutError:
            self.xknx.connection_manager.cemi_count_outgoing_error += 1
            raise ConfirmationError(
                f"L_DATA_CON Data Link Layer confirmation timed out for {cemi}"
            ) from None
        self.xknx.connection_manager.cemi_count_outgoing += 1

    def handle_raw_cemi(self, raw_cemi: bytes) -> None:
        """Parse and handle incoming raw CEMI Frames."""
        try:
            cemi = CEMIFrame.from_knx(raw_cemi)
        except CouldNotParseCEMI as cemi_parse_err:
            logger.warning("CEMI Frame failed to parse: %s", cemi_parse_err)
            self.xknx.connection_manager.cemi_count_incoming_error += 1
            return
        except UnsupportedCEMIMessage as unsupported_cemi_err:
            logger.info("CEMI not supported: %s", unsupported_cemi_err)
            self.xknx.connection_manager.cemi_count_incoming_error += 1
            return
        self.handle_cemi_frame(cemi)

    def handle_cemi_frame(self, cemi: CEMIFrame) -> None:
        """Handle incoming CEMI Frames."""
        if not isinstance(cemi.data, CEMILData):
            logger.debug("Ignoring incoming non-link-layer CEMI: %s", cemi)
            return

        if cemi.code is CEMIMessageCode.L_DATA_CON:
            # L_DATA_CON confirmation frame signals ready to send next telegram
            self._l_data_confirmation_event.set()
            logger.debug("Incoming CEMI confirmation: %s", cemi)
            return
        if cemi.code is CEMIMessageCode.L_DATA_REQ:
            # L_DATA_REQ frames should only be outgoing.
            logger.warning("Received unexpected L_DATA_REQ frame: %s", cemi)
            self.xknx.connection_manager.cemi_count_incoming_error += 1
            return
        logger.debug("Incoming CEMI: %s", cemi)
        self.xknx.connection_manager.cemi_count_incoming += 1

        _cemi_data_is_data_secure = is_data_secure(cemi.data)
        if self.data_secure is None:
            if _cemi_data_is_data_secure:
                data_secure_logger.debug(
                    "Received DataSecure encrypted CEMI frame but no keys for DataSecure are initialized: %s",
                    cemi,
                )
                self.handle_data_secure_key_issue(cemi.data, _cemi_data_is_data_secure)
                return
        else:
            try:
                cemi.data = self.data_secure.received_cemi(cemi_data=cemi.data)
            except DataSecureError as err:
                data_secure_logger.log(
                    err.log_level,
                    "Could not decrypt CEMI frame: %s",
                    err,
                )
                self.handle_data_secure_key_issue(cemi.data, _cemi_data_is_data_secure)
                return

        telegram = cemi.data.telegram()
        telegram.direction = TelegramDirection.INCOMING
        telegram.data_secure = _cemi_data_is_data_secure
        self.telegram_received(telegram)

    def telegram_received(self, telegram: Telegram) -> None:
        """Forward Telegram to upper layer."""
        if isinstance(telegram.tpci, tpci.TDataGroup):
            self.xknx.telegrams.put_nowait(telegram)
            return
        if (
            isinstance(telegram.destination_address, IndividualAddress)
            and telegram.destination_address != self.xknx.current_address
        ):
            return
        self.xknx.management.process(telegram)

    def handle_data_secure_key_issue(
        self, cemi_data: CEMILData, received_data_secure: bool
    ) -> None:
        """Handle DataSecure telegrams with missing or invalid keys."""
        self.xknx.connection_manager.undecoded_data_secure += 1
        if isinstance(cemi_data.tpci, tpci.TDataGroup):
            telegram = cemi_data.telegram()
            telegram.direction = TelegramDirection.INCOMING
            telegram.data_secure = received_data_secure
            self.xknx.telegram_queue.received_data_secure_group_key_issue(telegram)
